--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2016  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]


local TV=3.1;
local dbcver=1;									-- database communication version

if (not DuckMod) then DuckMod = {}; end			-- Make it if empty
if (not DuckMod[TV]) then						-- Don't mess with it if it's loaded
	DuckMod[TV]={};
end

local DM=DuckMod[TV];
local _;
local NATIVE="NativE";

globalDMROOT=DM;


local DUCKNET_TREESEP="\8";				-- Tree separator
local hiddenKey="_DMDB-";
local membersDB=hiddenKey.."share-guildlistdatabases";
local guildsDB=hiddenKey.."share-guildlists";
local pf="dm"..tostring(TV);
local interface=nil;
local pairsType=nil;

DM.Database={
	Separator=DUCKNET_TREESEP,
	DataVersion=1,
	LastStamp=0,			-- Last out of the database
	Redundant=nil,			-- Setting for written data
	BootDelay=15,
	StartBootDelay=15,
	PIPER_IDLE=0,			-- No attempt made
	PIPER_USED=1,			-- 1=used once, 2=twice, etc
	PIPER_UNAVAILABLE=-1,	-- Do not attempt this
	Com2={},
	API={},
	Util={ MT={}, },
	Info={ WhoList={}, },
	Addon={},
	Queue={},
	WQ={},
};
DM_PRUNE_NEVER=0;
DM_PRUNE_NORMAL=1;
DM_PRUNE_FIRST=10;			-- 11 second, 12 third, etc

DM.DatabaseQueue={};
DM.DatabaseData={};
dbPath={};						-- global place-holder
local dbd=DM.DatabaseData;		-- database data
local dbc=DM.Database;			-- database class
local api=dbc.API;
local util=dbc.Util;
local wq=dbc.WQ;
local gtime=time;


local function dbg(h,t)
	print(h,":");
	for _i,_d in pairs(t) do print(_i,_d); end
end
local function say(...)
	print("|cFF00FFFF",...);
end
local function saydate(a,b)
	say("Current:",date("%c",a),"and external:",date("%c",b));
end
local function addclass(class,container)
	for k,v in pairs(class) do container[k]=v; end end			-- add class to container
local function time()
	return gtime()+dbc.bgc.offset;
end

-- Set up for handling
function dbc:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-DB-Messager"..TV,UIParent);
		if (not self.TheFrame) then
			DM:Chat("Could not create DuckMod DB UI frame",1);
			return;
		end
		self.TheFrame:SetScript("OnEvent",dbc.OnEvent);
		self.TheFrame:SetScript("OnUpdate",dbc.Heartbeat);
		self.bgc={							-- background communication
			offset=nil, high=0, offset=0,
			iface=DM.Com2:CreateInterface(pf.."bgc",dbc.bgc),
		};
		dbc.bgc.iface:Send(gtime(),"timeme");
--	if (name=="timeme") then dbc.bgc.iface:Send(gtime(),"time"); return; end
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-DBWM-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then
			DM:Chat("Could not create DuckMod DB WM frame",1);
			return;
		end
--		self.TheFrameWorldMap:SetScript("OnEvent",dbc.OnEvent);
		self.TheFrameWorldMap:SetScript("OnUpdate",dbc.Heartbeat);
	end
end

function dbc.bgc(iface,data,name,meta,toon)
	if (name=="timeme") then data=gtime(); dbc.bgc.iface:Send(data,"time"); name="time"; end	-- reply and grab data
	if (name=="time") then
		if (data>dbc.bgc.high) then dbc.bgc.high=data; end
		dbc.bgc.offset=data-gtime();
	end
end

-- An event has been received
function dbc:OnEvent(event,arg1,arg2,arg3,arg4,arg5,arg6,_,_,arg9)
	-- Database inter-guild local roster update
	if (event=="GUILD_ROSTER_UPDATE") then
--print("GUILD_ROSTER_UPDATE");
		if (arg1) then return; end		-- Some change, so don't all start sync together
		local list={};
		local num=GetNumGuildMembers();
		for i=1,num do
			local name=GetGuildRosterInfo(i);
			list[name]=dbc.PIPER_IDLE;
		end
		local guild=GetGuildInfo("player");
		if (not guild) then return; end
--print(num.." members in "..guild);
		dbc.Info:UpdateGuildMemberList(guild,list);
	end

	-- Database inter-guild external roster update
	if (event=="WHO_LIST_UPDATE") then
print("WHO_LIST_UPDATE");
		local list={};
		local num=GetNumWhoResults();
		local guild;
		for i=1,num do
			_,guild=GetWhoInfo(i);
			if (not list[guild]) then list[guild]=1;
			else list[guild]=list[guild]+1; end
		end
		local goal,amount="",0;
		for guild,count in pairs(list) do
			if (count>amount) then goal=guild; amount=count; end
		end
		guild=goal;
		wipe(list);
		for i=1,num do
			local name,gname=GetWhoInfo(i);
			if (gname==guild) then list[name]=dbc.PIPER_IDLE; end
		end
		-- this function MUST be called, so can not abort between WHO_LIST_UPDATE and UpdateGuildMemberList
		dbc.Info:UpdateGuildMemberList(guild,list,true);	-- Merge it with the other list, if any
	end

	if (event=="COM2_DATA_RECEIVED") then
-- arg1 inter
-- arg2 build.da
-- arg3 build.na
-- arg4 build.md
-- arg5 build.api
-- arg6 sender
	end
end

-- find handle from path
local function getDB(path)
	if (not path:find("//",1,true)==1) then return nil; end
	if (path:sub(-1)=="/") then path=path:sub(1,-2); end		-- no trailing separators
	local list={strsplit("/",path:sub(3))};
	local listlen=#list;
	local p=dbd;
--	for i,v in ipairs(list) do
	local i,v=1;
	while(i<=listlen) do
		v=list[i];
		if (not p[v]) then _v=tonumber(v);
			if (not _v or not p[_v]) then		-- not found as string or number
				for j=i+1,listlen do v=v.."/"..list[j]; i=i+1; if (p[v]) then break; end end	-- searchforward slash versions
				if (not p[v]) then return nil; end	-- couldn't find it
			else v=_v; end
		end
		p=util:Handle(p[v]);
		if (d_type(p)~="database") then return nil; end
		i=i+1;
	end
	return (p~=dbd and p) or nil;	-- not root so use ref, or use nil
end

local function getDBKEY(path)
--print(" -->",path);
	if (path:sub(-1)=="/") then path=path:sub(1,-2); end		-- no trailing separators
--	local key; path,key=path:match("^(.+)/([^/]+)$"); if (not path) then return nil,nil; end	-- split at last slash, malformed
	local key,keypart,ref="";
	repeat
		if (key~="") then key="/"..key; end
		path,keypart=path:match("^(.+)/([^/]+)$"); if (not path) then return nil,nil; end	-- split at last slash, malformed
		key=keypart..key;
		if (path=="/") then ref=dbd; break; else ref=getDB(path); end
--print(" -->",path,key,ref);
	until (ref);
--	if (not ref) then return nil,nil; end									-- get ref, not found
	if (tonumber(key) and not ref._str_[key]) then key=tonumber(key); end	-- number in string, string key absent, use as number
	return ref,key;	-- return db ref and key
end

-- upgrade database before registration - so no meta
function util:UpgradeVersion(db)
	local ti,ty,re,da,meta;
	for k,v in pairs(db._str_) do
		meta="";
		ti,ty,re,da=v:match("^(%d+):(%a%a):([%+%-])(.*)$");			-- grab old format
		if (re=="-") then meta=meta.."r"; end						-- make meta
		if (ty=="st" and da:sub(1,2)=="03") then ty="ta"; end		-- convert flat tables
		db._str_[k]=ti..":"..ty..":"..meta..":"..da;				-- save new format: "time:type:meta:data"
		if (ty=="da") then util:UpgradeVersion(db[da]); end			-- do that database
	end
end

DM.DataPathEntryMetaTable={
	__index = function (t,k) return t.__path__.."/"..k; end,
}
-- The data distribution code
DM.DatabaseMetaTable={
	__index = function (t,k)
		if (k=="Path" or k=="File") then
			local path,tree="/",{strsplit(dbc.Separator,t._tree_)};
			for _,v in ipairs(tree) do path=path.."/"..v:sub(2); end
			if (k=="File") then local buf={ __path__=path }; setmetatable(buf,DM.DataPathEntryMetaTable); return buf; end
			return path;
		end
		if (rawget(t,"_lnk_")) then
			if (not t:DM_Get("PauseLinks")) then return util:Handle(rawget(t,"_lnk_"))[k]; end
			return nil;
		end
		dbc.LastStamp=0;
		-- if there's no data-key, return nil
		if (not t._str_[k]) then return nil; end
		local typ,meta,val;
--		dbc.LastStamp,typ,r,val=t._str_[k]:match("^(%d+):(%a%a):([%+%-])(.*)$");
		dbc.LastStamp,typ,meta,val=t._str_[k]:match("^(%d+):(%a%a):(.-):(.*)$");	-- time:type:meta:data
		dbc.LastStamp=tonumber(dbc.LastStamp);
		if (typ=="nu") then return tonumber(val);
		elseif (typ=="bo") then return (val=="true");
		elseif (typ=="ni") then return nil;
		elseif (typ=="da") then return rawget(t,val);	-- return real handle for database
		elseif (typ=="ta") then return unpacktable(val,meta:find("n",1,true));
		elseif (typ=="st") then if (meta:find("n",1,true)) then val=DM.Code:Unprotect(val,NATIVE); end
		end
		return val;		-- other
	end,
	__newindex = function (t,k,v)
		if (k=="Path" or k=="File") then return; end				-- read only
		local _rawget=rawget;
		if (_rawget(t,"_lnk_")) then
			if (not t:DM_Get("PauseLinks")) then util:Handle(_rawget(t,"_lnk_"))[k]=v; end
			return;
		end
		local nat=t:DM_Get("Native");								-- grab native setting
		if (util.SyncMeta) then nat=(util.SyncMeta:find("n",1,true) and true) or false; end	-- override active without native
--		if (nat==nil and t._tree_=="") then nat=false; end			-- not rooted (todatabase product), so native off
		local typ,tim=d_type(v),util.LockedTime or time();
		local ty,ref;
		if (t._str_[k]) then										-- there's something there
--			ty,ref=t._str_[k]:match("^%d+:(%a%a):[%+%-](.*)$");		-- get metainfo
			ty,ref=t._str_[k]:match("^%d+:(%a%a):.-:(.*)$");		-- time:type:meta:data
			if (ty~="da") then ref=nil; end							-- no ref here
		elseif (not util.LockedTime) then tim=tim-1000000000; end	-- Fake very old for new entries for other sync later
		local register=nil;
		if (typ=="table") then										-- a table has been provided
			if (not _rawget(t,"_endpoint_")) then typ="database";	-- override for table-to-database storage
			else v=packtable(v,nat); end							-- create packed table with native if applicable
		elseif (typ=="string") then
			if (nat) then v=DM.Code:Protect(v,NATIVE); end			-- apply native to strings
		end
		if (typ=="database") then									-- doing ref-write
			if (not ref) then ref=t:DM_UniqueKey(); end				-- must have ref
			v=todatabase(v);										-- if table - go database
			rawset(t,ref,v);										-- shove it directly in there
			register=true;
			typ=d_type(v);											-- update after forging
			v=ref;													-- set data to database name for save
		elseif (ref) then											-- doing normal write, and have old ref
			rawset(t,ref,nil);										-- overwrite with nil like a table would have done
		end
		-- save
		if (nat) then nat="n"; else nat=""; end
		t._str_[k]=tim..":"..typ:sub(1,2)..":"..(util.SyncMeta or nat)..":"..tostring(v);	-- time:type:meta:data
		util.SyncMeta=nil;
		if (register) then util:RegisterSubDatabases(t); end
		-- Written, so queue for send
		if (not util.LockedTime and ((_rawget(t,"_tree_") or "")~="")) then
			util:QueueForSend(nil,nil,_rawget(t,"_tree_"),k);
		end
	end,
	__tostring = function (t)
		return "database: "..util:GetLuaName(t._tree_);
	end,
	__type = "database",

-- metamethod "len" cannot be redefined. DM:Length(t) will handle it.
--	__len = function (t)
--		local count=0;
--		for _,v in pairs(t._str_) do if (v:match("^%d+:(%a%a):[%+%-].*$")~="ni") then count=count+1; end end
--		return count;
--	end,

	-- function dispatcher
	__call = function(t,index)
		index=index:lower();
		if (index=="event_queue_peek") then
			local q=rawget(util:GetRoot(t),"_queue_");
			return q[1];
		end
	end,

	-- insert new event
	__add = function(t,v)
		local function different(one,two)								-- check if two 1-level tables are in any unequal
			for k in pairs(one) do if (one[k]~=two[k]) then return true; end end
			for k in pairs(two) do if (one[k]~=two[k]) then return true; end end
			return nil;
		end
		local q=rawget(util:GetRoot(t),"_queue_");						-- get queue for this database
		local found;
		for i,e in ipairs(q) do if (not different(e,v)) then found=i; break; end end	-- grab old equal entry 
		if (found) then table.remove(q,found); end						-- remove it to simulate it being moved to the end
		table.insert(q,v);												-- add event to queue
		if (#q>1) then													-- more than one entry in queue
			local start=(q[1].busy and 2) or 1; found=nil;				-- skip nr 1 if it has busy set
			for i=start,#q do if (q[i].tree=="databaselist") then if (q[i].method=="send") then found=i; end end end	-- find a databaselist send
			if (found) then table.insert(q,q[found]); table.remove(q,found); found=nil; end						-- move it to the end of the list
			for i=start,#q do if (q[i].tree=="databaselist") then if (q[i].method=="poll") then found=i; end end end	-- find a databaselist poll
			if (found) then table.insert(q,q[found]); table.remove(q,found); end								-- move it to the end of the list
		end
		return t;
	end,
	-- remove a given number of events
	__sub = function(t,v)
		local q=rawget(util:GetRoot(t),"_queue_");
		while(v>0) do table.remove(q,1); v=v-1; end
		return t;
	end,
}

function dbc:Register(prefix,db,ef)
	local efn; if (ef) then efn=ef:GetName(); end
	self:Connect(db,prefix,efn or ef);
end

function dbc:Connect(db,prefix,ef)
	setmetatable(db,nil);							-- kill handlers
	local used=db._global_;
	if (not self.TheFrame) then dbc:Init(); end
	if (not db._global_) then db._global_={}; end
	if (not prefix) then if (not db._global_._prefix_) then return nil; end prefix=db._global_._prefix_; else db._global_._prefix_=prefix; end
	if (not ef) then ef=db._global_._eventframe_; else db._global_._eventframe_=ef; end
	if (type(ef)=="string") then ef=_G[ef]; end
	if (db._queue_) then wipe(db._queue_); db._queue_=nil; end
	DM.DatabaseQueue[prefix]={};
	DM.Talk[prefix]=time();
	db._queue_=DM.DatabaseQueue[prefix];		-- Set queue handler
	db._tree_=type(prefix):sub(1,1)..prefix;	-- Set root
	db._str_=db._str_ or {};
	db._prn_=db._prn_ or DM_PRUNE_NORMAL;
	db._global_.Limit=db._global_.Limit or 250;
	if (used and db._global_.Native==nil) then print("UpgradeVersion",prefix); util:UpgradeVersion(db); db._global_.Native=false; end
	local inter=DM.Com2:CreateInterface(prefix,dbc.Com2.RX,dbc.Com2.VC,dbc.Com2.DR,self.TheFrame,"GUILD");
	dbc.Addon[prefix]={ EventFrame=ef, dbL={}, ml={}, iface={ inter }, };
	dbd[prefix]=db;										-- Set reference
	util:SetInterface(db);								-- Set handlers
	util:SetDefaults(db);								-- Set all default globals
	util:RegisterSubDatabases(db);						-- add all nested databases
	local buf=db[guildsDB] or {};
	for sGuild,sForum in d_pairs(buf) do
		if (sForum) then db:DM_AddGuild(sGuild,sForum); end
	end
	-- all known interfaces registered, apply ban to all of them
	buf=unpacktable(db[hiddenKey.."SneezersTweezers"],true);
	for _,v in pairs(dbc.Addon[prefix].iface) do v:SetBan(buf); end
	-- run prune for the database
	DM.MT:Run("Pruning "..db.Path,util.MT.Prune,util.MT,prefix);	-- Fake call with self
	-- blind stuff for silent channels
--	db=db+{ method="EMPTY", };		-- add empty event to queue
	-- Set up to sync this database
--	self.Util:QueueForSend(prefix,"databaselist","",true);	-- Local only
	util:QueueForSend(nil,prefix,"databaselist");			-- everyone
	DM.QueueInited=true;
	return true;
end

function util:SetInterface(db)
	rawset(db,"DM_SetLink",api.SetLink);				-- ducklib.com
	rawset(db,"DM_Get",api.Get);						-- ducklib.com
	rawset(db,"DM_Set",api.Set);						-- ducklib.com
	rawset(db,"DM_SetRedundancy",api.SetRedundancy);	-- ducklib.com
	rawset(db,"DM_GetRedundancy",api.GetRedundancy);	-- ducklib.com
	rawset(db,"DM_PruneMode",api.PruneMode);			-- ducklib.com
	rawset(db,"DM_AddGuild",api.AddGuild);				-- ducklib.com
	rawset(db,"DM_RemoveGuild",api.RemoveGuild);		-- ducklib.com
	rawset(db,"DM_ListGuilds",api.ListGuilds);			-- ducklib.com
	rawset(db,"DM_UniqueKey",api.UniqueKey);			-- ducklib.com
	rawset(db,"DM_Registered",api.DatabaseRegistered);
	rawset(db,"DM_EndPoint",api.EndPoint);
	rawset(db,"DM_ProgressIB",api.ProgressIB);
	rawset(db,"DM_SetFrame",api.SetFrame);
	rawset(db,"DM_SetBan",api.SetBan);
	rawset(db,"DM_GetBan",api.GetBan);
	rawset(db,"DM_Property",api.Property);
	rawset(db,"DM_MetaIndex",api.MetaIndex);
	setmetatable(db,DM.DatabaseMetaTable);				-- Set handlers
end

-- Convert database to a normal table. This will of course remove any
-- information regarding prune, redundancy and time of storage.
function totable(t)
	local _d_type=d_type;
	if (_d_type(t)=="table") then return t; end				-- it's already a table
	if (_d_type(t)~="database") then return {t}; end		-- it's not a database, so enclose and return
	local buf={};											-- fresh table
	for k,v in d_pairs(t) do								-- walk all entries
		if (_d_type(v)=="database") then buf[k]=totable(v);	-- convert db to table
		else buf[k]=v; end									-- add entry as is
	end
	return buf;												-- return the new table
end

-- Convert table to a database. This will add default special entries, and
-- time of storage for all entries will be set to the current time.
function todatabase(t)
	local _d_type,_rawget=d_type,rawget;
	if (_d_type(t)=="database") then return t; end		-- it's already a database
	if (_d_type(t)~="table") then t={t}; end				-- it's not a table, so make one
	if (_rawget(t,"_tree_") and _rawget(t,"_str_") and _rawget(t,"_prn_")) then
		util:SetInterface(t);								-- set handlers
		return t;
	end
	local buf={
		_tree_="",									-- create dummy locator
		_str_={},									-- create storage
		_prn_=DM_PRUNE_NORMAL,						-- set normal pruning
	};
	util:SetInterface(buf);							-- set handlers
	for k,v in pairs(t) do							-- regular is fine here as it is a table
--		if (_d_type(v)=="table") then buf[k]=v;		-- create new level database from nested table
--		else buf[k]=v; end							-- add key with data to database
		buf[k]=v;									-- add through meta
	end
	return buf;										-- done this database
end

function api.SetFrame(db,eventframe)
	db._global_._eventframe_=eventframe;								-- shove
	dbc.Addon[db._global_._prefix_].EventFrame=eventframe;				-- shove
	if (not eventframe) then return; end								-- cleared, so done
	if (type(eventframe)=="string") then dbc.Addon[db._global_._prefix_].EventFrame=_G[eventframe];	-- grab current handle for this name
	else db._global_._eventframe_=eventframe:GetName(); end				-- grab name for this handle
end

function api.SetLink(db,dbName,linkDB)
	if (not linkDB) then rawset(db[dbName],"_lnk_",nil); return nil; end	-- clear link (if any)
	db[dbName]=todatabase({});								-- Make empty database
	rawset(db[dbName],"_lnk_",linkDB._tree_);				-- Set tree for external database
end

function api.Get(db,key)
	db=util:GetRoot(db);
	if (not db or not rawget(db,"_global_")) then return nil; end		-- Not a root db
	return db._global_[key];
end

function api.Set(db,key,value)
	db=util:GetRoot(db);
	if (not db or not rawget(db,"_global_")) then return; end				-- Not a root db
	if (key=="Native" and value==nil) then value=false; end		-- can't be nil
	db._global_[key]=value;
end

function api.EndPoint(db,value)
	local old=rawget(db,"_endpoint_");
	rawset(db,"_endpoint_",(value and true) or nil);
	local parent,pkey=util:GetLocationInfo(db._tree_);
	local a,meta,b=parent._str_[pkey]:match("^(%d+:%a%a:)(.-)(:.*)$");
	if (value and not meta:find("e",1,true)) then meta=meta.."e";				-- must add
	elseif (not value and meta:find("e",1,true)) then meta=meta:gsub("e","");	-- must remove
	else return; end															-- no change needed
	parent._str_[pkey]=a..meta..b;
end

function util:Handle(point,create)
	if (d_type(point)=="database") then												-- database supplied
		if (rawget(point,"_lnk_")) then return util:Handle(point._lnk_,create); end	-- I'm a link
		return point;																-- no link, so real database
	end
	if (create==true) then create=time(); end
	if (d_type(point)~="string") then return nil; end						-- not a tree, so error
	local prefix;				-- don't use self prefix as point may differ
	if (not point:match("^(%w-)"..dbc.Separator)) then prefix=point:sub(2);	-- root
	else prefix=(point:match("^(%w-)"..dbc.Separator)):sub(2); end			-- sub
	point={strsplit(dbc.Separator,point)};
	local goal=dbd;
	for num,entry in ipairs(point) do
		local loc=entry:sub(2); if (entry:sub(1,1)=="n") then loc=tonumber(loc); end
--		if (not goal[loc]) then
		if (d_type(goal[loc])~="database") then
			if (not create) then return nil; end
			if (goal._str_[loc]) then
--				local stamp=goal._str_[loc]:match("^(%d+):%a%a:[%+%-].*$"); stamp=tonumber(stamp);
				local stamp=goal._str_[loc]:match("^(%d+):%a%a:.-:.*$"); stamp=tonumber(stamp);
				if (stamp>create) then return nil; end						-- current entry is newer than incoming sub-entry
			end
			goal[loc]=todatabase({});
		end
		goal=goal[loc];
	end
	return goal;
end

function api.SetRedundancy(db,key)
	db=util:Handle(db);								-- in case of link
--	local ls,typ,val=db._str_[key]:match("^(%d+):(%a%a):[%+%-](.*)$");
	local a,meta,b=db._str_[key]:match("^(%d+:%a%a:)(.-)(:.*)$");
	if (meta:find("r",1,true)) then return; end		-- nothing to do
	util.SyncMeta=meta.."r";						-- meta storage override
	db[key]=db[key];								-- sync (also clears syncmeta)
end

function api.GetRedundancy(db,key)
	db=util:Handle(db);								-- in case of link
--	local r=db._str_[key]:match("^%d+:%a%a:([%+%-]).*$");
	local meta=db._str_[key]:match("^%d+:%a%a:(.-):.*$");
	return meta:find("r",1,true);
end

function api.Property(db,key,property,value)
	return db:DM_MetaIndex(key,"p",property,value);
end

function api.MetaIndex(db,key,indicator,index,value,metaonly)
	if (type(index)~="number" or index<1 or index>100) then return; end		-- wrong type or OOB
	if (type(indicator)~="string" or not indicator:match("^%l$")) then return; end	-- indicator is other than a single lower case letter
	indicator=string.format(indicator.."%02d",index-1);					-- create indicator or character + index number
	db=util:Handle(db);													-- in case of link
	local a,meta,b=db._str_[key]:match("^(%d+:%a%a:)(.-)(:.*)$");		-- grab beginning, meta, end
	if (type(value)~="number") then										-- not a number, so read it
		value=meta:match(indicator.."(%d+)");							-- grab value
		return tonumber(value);											-- return as number
	end
	if (meta:find(indicator,1,true)) then								-- check for existing indicator and index
		local a_,b_=meta:match("^(.*)"..indicator.."%d+(.*)");			-- grab all other stuff
		meta=a_..b_;													-- remake
	end
	meta=meta..indicator..value;
	if (metaonly) then return meta; end
	util.SyncMeta=meta;													-- meta storage override
	db[key]=db[key];													-- sync (also clears syncmeta)
end

function api.PruneMode(db,mode)
	db=util:Handle(db);								-- in case of link
	if (not db) then return; end
	if (mode) then rawset(db,"_prn_",mode); end
	return rawget(db,"_prn_");
end

function api.AddGuild(db,guild,forum,mode)
	local myGuild=GetGuildInfo("player"); if (not guild or guild=="" or guild==myGuild) then return; end
	db=util:GetRoot(db);												-- rewind tree
	local prefix=db:DM_Get("_prefix_");									-- get unique prefix
	if (not mode) then mode="NORMAL"; end								-- no exceptions
	dbc.Info.WhoList[guild]=true;										-- flag for member retrieval
	if (not db[membersDB]) then db[membersDB]={}; db[membersDB]:DM_EndPoint(true); end	-- prep list
	if (d_type(db[guildsDB])~="database") then db[guildsDB]={}; db[guildsDB]:DM_EndPoint(true); db[guildsDB]:DM_PruneMode(DM_PRUNE_NEVER); end		-- prep list
	-- Set shareable guild member list database
	if (not db[membersDB][guild]) then db[membersDB][guild]={}; end		-- create member-list if missing
	db[guildsDB][guild]=forum;											-- set guild and share-name in registry
	-- we obviously have someone to share with, so force a roster update to make
	-- sure the other guild(s) gets an updated list of our own members.
	if (not dbc.Addon[prefix].OwnGuildRosterDone) then
		GuildRoster();
		util:QueueForSend(nil,nil,dbd[prefix]._tree_,myGuild);	-- Send to everyone
		dbc.Addon[prefix].OwnGuildRosterDone=true;
	end
	-- create communications interface for external guilds
	local inter=DM.Com2:CreateInterface(prefix,dbc.Com2.RX,dbc.Com2.VC,dbc.Com2.DR,ef,"WHISPER");
	table.insert(dbc.Addon[prefix].iface,inter);						-- add it to addon's interface list
	inter:SetReceiver(db[membersDB][guild],guild);						-- set list in interface
end


-- membersDB (_DMDB-share-guildlistdatabases)
--     guild 1
--         toon 1 =
--         toon n =
--     guild n
--         toon n =

-- guildsDB (_DMDB-share-guildlists)
--     guild 1 = forum 1
--     guild n = forum n

function api.RemoveGuild(db,guild)
	if (not guild or guild=="") then return; end
	db=util:GetRoot(db);
	local prefix=db:DM_Get("_prefix_");
	DM.Database.Info.WhoList[guild]=nil;

	local dbname;
	dbname=membersDB;
	if (db[dbname]) then db[dbname][guild]=nil; end
	dbname=guildsDB;
	if (db[dbname]) then db[dbname][guild]=nil; end
end

function api.ListGuilds(db)
	db=util:GetRoot(db);
	if (not db[guildsDB]) then return {}; end
	local list={};
	for g,gd in pairs(db[guildsDB]) do if (gd) then table.insert(list,g); end end
	return list;
end

function api.UniqueKey(db)
	while(true) do
		local key="auk_"..string.format("%X",tostring(math.random(0x7FFFFFFF))).."_auk";
		if (not db._str_[key] and not rawget(db,key)) then return key; end
	end
end

function api.DatabaseRegistered(db)
	return true;
end

function api.ProgressIB(db)
	return dbc.Addon[db:DM_Get("_prefix_")].iface[1]:ProgressIB();
end

function api.SetBan(db,toon,reason)
	db=util:GetRoot(db); if (not db) then return; end
	local list=unpacktable(db[hiddenKey.."SneezersTweezers"],true);
	list[toon]=reason;
	for _,v in pairs(dbc.Addon[db._global_._prefix_].iface) do v:SetBan(list); end		-- shove new list
	db[hiddenKey.."SneezersTweezers"]=packtable(list,true);
end

function api.GetBan(db,toon)
	db=util:GetRoot(db); if (not db) then return; end
	return unpacktable(db[hiddenKey.."SneezersTweezers"],true)[toon];
end

function util:SetDefaults(db)
	if (not rawget(db,"_global_")) then return; end		-- Not a root db
	for k,v in pairs(db._global_) do
		if (d_type(k)=="string") then
			local key=k:match("^(.+)_Default$");
			if (key) then db:DM_Set(key,v); end
		end
	end
end

function util:GetRoot(db)
	if (not db._tree_) then return nil; end
	db=util:Handle(db);		-- resolve current link if any
	if (not db._tree_) then return nil; end
	local prefix;
	if (not db._tree_:match("^(%w-)"..dbc.Separator)) then prefix=db._tree_:sub(2);	-- root
	else prefix=(db._tree_:match("^(%w-)"..dbc.Separator)):sub(2); end
	return dbd[prefix];
end

function util:SendDatabaseList(inter,prefix)
	local list={};
	util:AddDatabaseToList(dbd[prefix],list);
--	MessageBoard_LS.DatalistOut=DM:CopyTable(list);
	util:InterSend(inter,prefix,list,UnitName("player"),{ cont="dbl" },nil);		-- not local only
end

--
--	Sending interface for databases. All registered guilds are cycled.
--
--		util:InterSend(inter,prefix,value,nil,{ cont="val", key=key, tree=tree, ver=dbc.LastStamp },nil);		-- not local only
--		util:InterSend(inter,prefix,handle._str_,nil,{ cont="db", tree=tree, ver=dbc.LastStamp },nil);		-- not local only
function util:InterSend(inter,prefix,data,name,meta,localonly)
--if (meta and meta.cont=="dbl") then
--  say(prefix,"Sending my own databaselist.");
--elseif (meta and meta.tree) then
--  say(prefix,"Sending:",util:GetLuaName(meta.tree,meta.key),(meta.cont and meta.cont) or "with no contents descriptor.");
--else
--  say("Sending with no tree for",prefix);
--end

	meta=meta or {}; meta.dbcver=dbcver;
	if (inter) then inter:Send(data,name,meta); return; end				-- specified interface only
	for _,inter in pairs(dbc.Addon[prefix].iface) do
		inter:Send(data,name,meta);
		if (localonly) then return; end		-- guild share is index 1
	end
end

--
--	Requesting interface for databases. All registered guilds are cycled.
--
function util:InterRequest(inter,prefix,ver,name,meta,localonly)
	meta=meta or {}; meta.dbcver=dbcver;
	if (inter) then inter:Request(name,meta,ver); return; end			-- specified interface only
	for _,inter in pairs(dbc.Addon[prefix].iface) do
		inter:Request(name,meta,ver);
		if (localonly) then return; end		-- guild share is index 1
	end
end

function util:AddDatabaseToList(db,list)
	if (db._tree_=="") then print("Empty tree detected in AddDatabaseToList()"); return; end
	list[db._tree_]=self:GetIDForDatabase(db);		-- add yourself
	for _,dbTable in d_pairs(db,true) do			-- add nested databases without following links
		if (d_type(dbTable)=="database") then self:AddDatabaseToList(dbTable,list); end
	end
--	Save created database list for MessageBoard
--	if (MessageBoard_DB) then
--		local p;
--		if (not MessageBoard_DB.dbLists) then MessageBoard_DB.dbLists={}; end
--		p=MessageBoard_DB.dbLists;
--		if (not p[UnitName("player")]) then p[UnitName("player")]={}; end
--		p=p[UnitName("player")];
--		p[time()]=DM:CopyTable(list);
--	end
end

-- _str_ and _prn_ already be set by creation or load
function util:RegisterSubDatabases_(parent,key,db)
	rawset(db,"_tree_",parent._tree_..dbc.Separator..d_type(key):sub(1,1)..key);
	self:SetInterface(db);
	for k,v in pairs(db._str_) do
--		if (v:match("^%d+:(%a%a):[%+%-].*$")=="da") then
		if (v:match("^%d+:(%a%a):.-:.*$")=="da") then
			self:RegisterSubDatabases_(db,k,db[k]);
		end
	end
end

function util:RegisterSubDatabases(db)
	if (d_type(db)~="database") then return; end	-- can't work with this
	for k,v in pairs(db._str_) do
--		if (v:match("^%d+:(%a%a):[%+%-].*$")=="da") then
		if (v:match("^%d+:(%a%a):.-:.*$")=="da") then
			self:RegisterSubDatabases_(db,k,db[k]);
		end
	end
end

-- This does NOT send nested databases
function util:SendValue(inter,prefix,tree,key)
	local datameta;
	-- Database list
	if (tree=="databaselist") then
		util:SendDatabaseList(inter,prefix);			-- OK, and poll it to get it going
		util:QueueForPoll(inter,prefix,"databaselist",UnitName("player"),0);
		return true;
	end

	-- Single values
	if (key) then
		local handle=util:Handle(tree);
		if (not handle) then return true; end	-- fake valid transmission to remove this abomination
		local value=handle[key];
		datameta=handle._str_[key]:match("^%d+:%a%a:(.-):.*$");
		if (d_type(value)=="database") then return self:SendValue(inter,prefix,rawget(value,"_tree_"),nil); end
		util:InterSend(inter,prefix,value,nil,{ cont="val", key=key, tree=tree, ver=dbc.LastStamp, meta=datameta },nil);	-- not local only
		return true;
	end

	-- Complete databases
	local handle=util:Handle(tree);
	local parent,pkey,_,stamp=self:GetLocationInfo(tree);
	if (parent) then
		datameta=parent._str_[pkey]:match("^%d+:%a%a:(.-):.*$");
	end
	util:InterSend(inter,prefix,handle._str_,nil,{ cont="db", tree=tree, ver=stamp, meta=datameta },nil);	-- not local only
	return true;
end

-- accepts both databases and trees
function util:GetLocationInfo(tree,create)
--	tree=util:Handle(tree); if (not tree._tree_) then return nil; end		-- resolve current link if any
	tree=util:Handle(tree,create); if (not tree or not tree._tree_) then return nil; end		-- resolve current link if any
	tree=tree._tree_;
	if (not tree:find(dbc.Separator)) then return nil,"root","database",0; end	-- root - don't mess with it
	local parent,typ,name=tree:match("^(.+)"..dbc.Separator.."(%a)(.-)$");	-- last separator
	if (typ=="n") then name=tonumber(name); end
	parent=util:Handle(parent);		-- get parent
	if (not parent) then return "unreachable"; end
--	local stamp=tonumber(parent._str_[name]:match("^(%d+):%a%a:[%+%-].*$"));
	local stamp=tonumber(parent._str_[name]:match("^(%d+):%a%a:.-:.*$"));
	return parent,name,d_type(parent[name]),stamp;	-- parentDB,key,type,stamp
end

function dbc.Com2.ReDirect(inter,meta)
	if (meta and meta.forum) then
		for id,db in pairs(dbd) do											-- walk all known databases
			local buf=db[guildsDB] or {};									-- grab forum list for this database
			for sg,gForum in d_pairs(buf) do								-- walk all its forum names
				if (gForum==meta.forum and (db[membersDB][sg] or {})[sender]) then	-- if a forum matches and member is in that guild
					return id,db;											-- found goal database ad 'db' with prefix 'id'
				end
			end
		end
	end
	return inter.ID,dbd[inter.ID];
end

--	t.cbRX=cbRX;	-- data RX
--		cbRX(interface,data,data name,metadata,sender toon,api used);
function dbc.Com2.RX(inter,data,name,meta,sender,api)
	local pf,db=dbc.Com2.ReDirect(inter,meta);
	if (not DM.DatabaseQueue[pf]) then return; end		-- Quick way out
	if (not db) then return; end						-- Quick way out
	if (not meta or meta.dbcver~=dbcver) then return; end

	if (meta.cont=="val") then						-- incoming value
		if (not meta.tree) then return; end				-- all types must have a tree
		if (not meta.key) then return; end;			-- no key
--say(pf,": Received value for",util:GetLuaName(meta.tree,meta.key));
		local goal=util:Handle(meta.tree,meta.key);	-- Create it if doesn't exist or is older
		if (not goal) then return; end
		local newdata=nil; if (not goal._str_[meta.key]) then newdata=true; end
		_=goal[meta.key];							-- Update stamp
		if (dbc.LastStamp>meta.ver) then
--say("My current data is newer, so aborting save.");
--saydate(dbc.LastStamp,meta.ver);
			return; end	-- mine's better
		-- check for guild member-lists and discard own guild listing
		if (meta.tree:match(membersDB.."$")) then	-- check for this as an ending
			local myGuild=GetGuildInfo("player");
			if (meta.key==myGuild) then return; end	-- don't accept own guild at all
		end
		-- normal storage
		util.LockedTime=meta.ver;					-- Override time
		util.SyncMeta=meta.meta;
		goal[meta.key]=data;						-- Set value
		util.LockedTime=nil;						-- Re-enable time
--say("Saved.");
		if (not tostring(meta.key):match("^"..hiddenKey)) then		-- not a hidden share
			util:SendEvent(pf,"DMEVENT_DATABASE_SYNC",goal,meta.key,newdata);
		end
		if (inter.Channel:upper()=="WHISPER" and inter.Receiver) then	-- received from external source
--say("Private reception. Queuing for guild broadcast.");
			util:QueueForSend(nil,pf,meta.tree,meta.key,true);	-- distribute locally
		end
		return;
	elseif (meta.cont=="db") then					-- incoming database
		if (not meta.tree) then return; end				-- all types must have a tree
--say(pf,": Received database",util:GetLuaName(meta.tree));
		local parent,pkey=util:GetLocationInfo(meta.tree,meta.ver);	-- create empty database if location is empty or older
		if (not parent and not pkey) then
--say("ERROR: Anon database without parent detected.");
		return; end	-- no parent, can't save (do not create it)
		if (parent) then
			if (parent._str_[pkey]) then
				_=parent[pkey];						-- update stamp reading and save it for single access
				if (d_type(_)~="database") then
					if (meta.ver<=dbc.LastStamp) then
--say("Current non-db data is newer than incoming database.");
--saydate(dbc.LastStamp,meta.ver);		-- first current, second incoming/external
					return; end	-- not newer received
					util.LockedTime=meta.ver; parent[pkey]={}; util.LockedTime=nil;
--say("Current older non-db data was overwritten by empty db at",date("%c",meta.ver));
				end
			else	-- key does not exist
				util.LockedTime=meta.ver; parent[pkey]={}; util.LockedTime=nil;
--say("Empty nested database created at",date("%c",meta.ver));
			end
			-- handle meta for incoming database entry
			local a,b=parent._str_[pkey]:match("^(%d+:%a%a:).-(:.*)$");			-- extract info
			parent._str_[pkey]=a..meta.meta..b;									-- shove meta
			if (meta.meta:find("e",1,true)) then parent[pkey]:DM_EndPoint(true); end	-- set-up by meta
		elseif (pkey~="root") then
--say("ERROR: Non-root database without parent detected.");
		return; end		-- some other problem
		-- merge entries
		local handle=util:Handle(meta.tree);
--say("Merging incoming database",util:GetLuaName(meta.tree));
		for entry,eData in pairs(data) do			-- data is remote raw _str_ table
			local newdata,savedata=nil,nil;
--			local sNew,tNew=eData:match("^(%d+):(%a%a):[%+%-].*$"); sNew=tonumber(sNew);
			local sNew,tNew,mNew=eData:match("^(%d+):(%a%a):(.-):.*$"); sNew=tonumber(sNew);
			if (not handle._str_[entry]) then				--==>> it's new data
				newdata=true; savedata=true;
			else
--				local sOld,dOld,tOld=handle._str_[entry]:match("^(%d+)(:(%a%a):[%+%-].*)$"); sOld=tonumber(sOld);
				local sOld,a,tOld,b=handle._str_[entry]:match("^(%d+)(:(%a%a):).-(:.*)$"); sOld=tonumber(sOld);
				if (tOld~="da" or tNew~="da") then			--==>> one is not db
					if (sNew>sOld) then savedata=true;
--say("Existing type",tOld,"will be overwritten. - Updated stamp:");
--saydate(sOld,sNew);		-- first current, second incoming/external
					end	--==>> do normal evaluation
				elseif (sNew>sOld) then						--==>> both are databases
					rawset(handle._str_,entry,sNew..a..mNew..b);	-- update stamp and meta
--say("Existing and incoming nested database detected:",util:GetLuaName(handle,entry),"- Updated stamp:");
--saydate(sOld,sNew);		-- first current, second incoming/external
				end
			end
			if (savedata) then
				if (tNew=="da") then						-- a database is incoming
					util.LockedTime=sNew;					-- set lock
					util.SyncMeta=mNew;						-- force meta
					handle[entry]={};						-- overwrite/save by meta
					util.LockedTime=nil;					-- clear lock
--say("Creating empty nested database at",util:GetLuaName(handle,entry),"at",date("%c",sNew));
				else										-- non-database incoming
					handle._str_[entry]=eData;				-- store raw non-database entry
--say("Storing raw "..tNew.."-type data at",util:GetLuaName(handle,entry));
				end
				util:SendEvent(pf,"DMEVENT_DATABASE_SYNC",handle,entry,newdata);	-- updated
			end
		end
		if (inter.Channel:upper()=="WHISPER" and inter.Receiver) then	-- received from external source
--say("Private reception. Queuing for guild broadcast.");
			util:QueueForSend(nil,pf,meta.tree,meta.key,true);	-- distribute locally
		end
		return;
	elseif (meta.cont=="dbl") then					-- incoming database list
		if (not DM.Database.Addon[pf].dbL[name]) then DM.Database.Addon[pf].dbL[name]={}; end
		wipe(DM.Database.Addon[pf].dbL[name]);	-- better for garbagecollect to wipe first
		DM.Database.Addon[pf].dbL[name]=DM:CopyTable(data,DM.Database.Addon[pf].dbL[name]);
--say("Databaselist for",name,"received. Stored. No other action taken.");
		-- no need for local sync of incoming external databaselist
		return;
	end
end

--	t.cbVC=cbVC;	-- data version check typically for negotiation
--		cbVC(interface,data name,metadata);
function dbc.Com2.VC(inter,name,meta)
	local pf,db=dbc.Com2.ReDirect(inter,meta);
	if (not DM.DatabaseQueue[pf]) then return; end		-- Quick way out
	if (not db) then return; end						-- Quick way out

	-- A poll for value stamp
	if (meta.key and meta.tree) then
--say("Someone is checking my version of value",util:GetLuaName(meta.tree,meta.key));
		local handle=util:Handle(meta.tree); if (not handle) then return 0; end
--		return tonumber(handle._str_[meta.key]:match("^(%d+):%a%a:[%+%-].*$"));
		return tonumber(handle._str_[meta.key]:match("^(%d+):%a%a:.-:.*$"));
	end

	-- A poll for database ID/version/stamp
	if (not meta.key and meta.tree and meta.ve) then
		local handle=util:Handle(meta.tree); if (not handle) then
--say("Someone is checking my version for database",util:GetLuaName(meta.tree),"=> Don't have it.");
		return 0; end
		if (util:GetIDForDatabase(handle)==meta.ve) then
--say("Someone is checking my version for database",util:GetLuaName(meta.tree),"=> I have the same version.");
		return 0; end
--say("Someone is checking my version for database",util:GetLuaName(meta.tree),"=> Mine is different.");
		return 1;
	end

	-- A poll for database list ID (ID=toon=meta.key)
	if (meta.key and meta.dblist) then
		if (dbc.Addon[pf].dbL[meta.key]) then
			local list={};
			util:AddDatabaseToList(dbd[pf],list);				-- Make my own database list
			-- Remove same and non-existent based on own database
			for tree,id in pairs(dbc.Addon[pf].dbL[meta.key]) do	-- cycle external list
				if (list[tree] and list[tree]==id) then list[tree]=nil; end	-- Both have same
			end
			-- Walk it
			local stamper=0;
			for _ in pairs(list) do stamper=stamper+1; end
			return stamper;
		end
--say("Someone is checking databasedlist version for a missing list from user",meta.key);
		return 0;
	end

	-- default
--say("Someone is making an illegal version check:");
--dbg("meta",meta);
	return 0;
end

--	t.cbDR=cbDR;	-- a request for data to be sent (typ. from a negotiation)
--		cbDR(interface,data name,metadata);
function dbc.Com2.DR(inter,name,meta)
	local pf,db=dbc.Com2.ReDirect(inter,meta);
	if (not DM.DatabaseQueue[pf]) then return; end		-- Quick way out
	if (not db) then return; end						-- Quick way out
	if (not meta or meta.dbcver~=dbcver) then return; end

--dbg(pf..": DR called",meta);
--print(dbc.Addon[pf]);
	-- databaselist
	if (meta.key and meta.dblist) then									-- it's a database list
		if (not dbc.Addon[pf] or not dbc.Addon[pf].dbL[meta.key]) then return; end	-- i don't have any such stuff
--say("I need to queue databaselist differences for",meta.key,pf);
		local list={};
		util:AddDatabaseToList(db,list);								-- create my own database list
		if (DM:CompareTables(list,dbc.Addon[pf].dbL[meta.key])) then return; end	-- they're equal, so nothing to do
		for tree,id in pairs(dbc.Addon[pf].dbL[meta.key]) do		-- cycle external list
			if (list[tree] and list[tree]==id) then						-- i have the same entry
				list[tree]=nil;											-- so remove it from my list
			end
		end

--say("Differences for",meta.key,pf,"database are:");
		for tree in pairs(list) do util:QueueForSend(nil,pf,tree);	-- queue all databases that are different for send
--say("==>",util:GetLuaName(tree));
		end	-- all non-equal databases
		util:QueueForSend(nil,pf,"databaselist");				-- send own list when all done
		return;
	end

	-- database and value
--say(pf,": Generic request for",util:GetLuaName(meta.tree,meta.key),"queued for output.");
	util:QueueForSend(nil,pf,meta.tree,meta.key,nil);
end

-- get one-level unique(ish) ID for a database
function util:GetIDForDatabase(db)
	if (not db) then return 0; end								-- nothing i can use
	local id=0;													-- new id
	for _,eData in pairs(db._str_) do							-- walk storage
--		id=id+tonumber(eData:match("^(%d+):%a%a:[%+%-].*$"));	-- add stamp
		id=id+tonumber(eData:match("^(%d+):%a%a:.-:.*$"));
	end
	id=id+1;													-- id=0 is a special case, so avoid that
	return id;
end

local function getprefix(prefix,tree)
	if (prefix and prefix~="") then return prefix; end
	if (not tree:match("^(%w-)"..dbc.Separator)) then return tree:sub(2); end	-- root
	return (tree:match("^(%w-)"..dbc.Separator)):sub(2);						-- nested
end


--function DM.Database.Util:QueueForSend(prefix,tree,key,noguild)
function util:QueueForSend(inter,prefix,tree,key,localonly)
	if (tree=="") then return; end			-- skip unrooted databases (typically todatabase)
	prefix=getprefix(prefix,tree);			-- recreate prefix if not specified
--say(prefix,"Queue for send",util:GetLuaName(tree,key),(localonly and "=> local") or "global");
	DM.Talk[prefix]=time();					-- Delay further
	local q={ method="send", tree=tree, key=key, inter=inter };						-- basic metadata
	if (tree=="databaselist") then q.offset=20; end						-- additional wait
	q.localonly=localonly;
	dbd[prefix]=dbd[prefix]+q;				-- add event to queue
end

--function DM.Database.Util:QueueForPoll(prefix,tree,key,stamp,noguild)
function util:QueueForPoll(inter,prefix,tree,key,stamp)
	if (tree=="") then return; end			-- skip unrooted databases (typically todatabase)
	prefix=getprefix(prefix,tree);			-- recreate prefix if not specified
--say(prefix,"Queue for poll",util:GetLuaName(tree,key));
	DM.Talk[prefix]=time();					-- Delay further
	local q={ method="poll", tree=tree, key=key, inter=inter, instamp=stamp };
	dbd[prefix]=dbd[prefix]+q;
end

function listType(typ)
	pairsType=typ;
end

function d_next(t,k)
	if (d_type(t)~="database") then return next(t,k); end
	local kv=nil;
	repeat repeat
			k=next(t._str_,k);
			if (k) then kv=t[k]; end
		until (not tostring(k):match("^"..hiddenKey));	-- get next if this is a database zero (valid key, nil data)
	until (not k or kv);			-- Invoke metatable for data
	return k,kv;
end

-- Support for metatable types
function d_type(data)
	local m=getmetatable(data);
	if (m and m.__type) then return m.__type; end
	return type(data);
end

function table.dmsort(t,func)
	local buf={};
	for k,d in pairs(t._str_) do table.insert(buf,{k=k,d=d}); end
	table.sort(buf,func);
	return buf;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function d_pairs_up(t,nl)
	if (d_type(t)~="database") then return pairs(t); end		-- fallback
	if (rawget(t,"_lnk_")) then
		if (t:DM_Get("PauseLinks")) then nl=true; end
		if (nl) then return function() return nil; end,t,nil; end		-- empty iterator
		t=util:Handle(rawget(t,"_lnk_"));
		if (not t) then return function() return nil; end,t,nil; end	-- empty iterator
	end
	local oi=table.dmsort(t,
				function(A,B)
--					if (tonumber(A.d:match("^(%d+):%a%a:[%+%-].*$"))<tonumber(B.d:match("^(%d+):%a%a:[%+%-].*$"))) then return true; end
					if (tonumber(A.d:match("^(%d+):%a%a:.-:.*$"))<tonumber(B.d:match("^(%d+):%a%a:.-:.*$"))) then return true; end
				end
			);
	local types=pairsType; pairsType=nil;
	return function (t,k)
		local kv;
		repeat
			kv=nil; k=table.remove(oi,1);							-- pull top one and get its data
			if (not k) then return nil,nil; end						-- list empty
			if (k.k) then kv=t[k.k]; end							-- invoke metatable for data
			if (types and types~=d_type(kv)) then kv=nil; end		-- we don't want that data type
		until (kv and not tostring(k.k):match("^"..hiddenKey));		-- until there's data and not hidden key
		return k.k,kv;
	end,t,nil;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function d_pairs_dn(t,nl)
	if (d_type(t)~="database") then return pairs(t); end		-- fallback
	if (rawget(t,"_lnk_")) then
		if (t:DM_Get("PauseLinks")) then nl=true; end
		if (nl) then return function() return nil; end,t,nil; end		-- empty iterator
		t=util:Handle(rawget(t,"_lnk_"));
		if (not t) then return function() return nil; end,t,nil; end	-- empty iterator
	end
	local oi=table.dmsort(t,
				function(A,B)
--					if (tonumber(A.d:match("^(%d+):%a%a:[%+%-].*$"))>tonumber(B.d:match("^(%d+):%a%a:[%+%-].*$"))) then return true; end
					if (tonumber(A.d:match("^(%d+):%a%a:.-:.*$"))>tonumber(B.d:match("^(%d+):%a%a:.-:.*$"))) then return true; end
				end
			);
	local types=pairsType; pairsType=nil;
	return function (t,k)
		local kv;
		repeat
			kv=nil; k=table.remove(oi,1);							-- pull top one and get its data
			if (not k) then return nil,nil; end						-- list empty
			if (k.k) then kv=t[k.k]; end							-- invoke metatable for data
			if (types and types~=d_type(kv)) then kv=nil; end		-- we don't want that data type
		until (kv and not tostring(k.k):match("^"..hiddenKey));		-- until there's data and not hidden key
		return k.k,kv;
	end,t,nil;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function d_pairs(t,nl)
	if (d_type(t)~="database") then return pairs(t); end
	if (rawget(t,"_lnk_")) then
		if (t:DM_Get("PauseLinks")) then nl=true; end
		if (nl) then return function() return nil; end,t,nil; end		-- empty iterator
		t=util:Handle(rawget(t,"_lnk_"));
		if (not t) then return function() return nil; end,t,nil; end	-- empty iterator
	end
	local types=pairsType; pairsType=nil;
	return function (t,k)
		local kv;
		repeat
			k=next(t._str_,k);
			if (not k) then return nil,nil; end						-- no more
			kv=t[k];												-- invoke metatable for data
			if (types and types~=d_type(kv)) then kv=nil; end		-- we don't want that
		until (kv and not tostring(k):match("^"..hiddenKey));		-- until there's data and not hidden key
		return k,kv;
	end,t,nil;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function d_ipairs(t,nl)
	if (d_type(t)~="database") then return ipairs(t); end
	if (rawget(t,"_lnk_")) then
		if (t:DM_Get("PauseLinks")) then nl=true; end
		if (nl) then return function() return nil; end,t,0; end		-- empty iterator
		t=util:Handle(rawget(t,"_lnk_"));
		if (not t) then return function() return nil; end,t,0; end	-- empty iterator
	end
	return function (t,k)
		k=k+1;
		if (t._str_[k]==nil) then return nil,nil; end
		return k,t[k];			-- Invoke metatable for data
	end,t,0;
end

--local function dbPF(comprefix,gettoon)
--	if (not DM.Net.DB[comprefix].Receiver) then return comprefix; end
--	local dbprefix,toon=comprefix:match("^(.+)"..DM.Net.Split1.."(.-)$");
--	if (not dbprefix or not toon) then return; end
--	if (gettoon) then return toon; end			-- to have only single-value returns
--	return dbprefix;
--end

--function dbc.Info:GetOnlineGuildMembers(guild)
--	SetWhoToUI(1);
--	FriendsFrame:UnregisterEvent("WHO_LIST_UPDATE");
--	SendWho("g-\""..guild.."\"");
--end

function util.MT:UpdateGuildLists()
	if (util.UpdatingGuildLists) then return; end
	util.UpdatingGuildLists="";

	SetWhoToUI(1);
	FriendsFrame:UnregisterEvent("WHO_LIST_UPDATE");
	repeat
		local guild,state=next(dbc.Info.WhoList);
		if (not guild or not state) then break; end		-- list is now empty
		local retry=0;
		repeat
			retry=retry+1;
			SendWho("g-\""..guild.."\"");		-- request these members
			local now=debugprofilestop();
			repeat DM.MT:Yield(true); until (debugprofilestop()-now>=5000 or util.UpdatingGuildLists==guild);
		until (util.UpdatingGuildLists==guild or retry>=3);		-- keep going until guild has been received or retried too many times
		if (util.UpdatingGuildLists==guild) then						-- got this guild
			-- set new list to interface here
			for prefix,db in pairs(dbd) do								-- do all prefixes
				for _,inter in pairs(dbc.Addon[prefix].iface) do		-- do all interfaces for this prefix
					if (inter.ReceiverListName==guild) then				-- there's an interface for this guild at this prefix
						inter:SetReceiver(db[membersDB][guild],guild);	-- so plug list in there
					end
				end
			end
		else
			dbc.Info.WhoList[guild]=nil;								-- most likely no members online
		end
	until (false);		-- forever
	FriendsFrame:RegisterEvent("WHO_LIST_UPDATE");

	util.UpdatingGuildLists=nil;
end

function dbc.Info:UpdateGuildMemberList(guild,list,merge)
	for prefix,db in pairs(dbd) do
--		DM.Script:SetValue(prefix,"UEGwait",-1);	-- tell potential scripts that a list is here
		if (db[membersDB]) then
			local oldlist,changed,valid=db[membersDB][guild],false,true;	-- select guild
			if (oldlist) then												-- old list exists
				for toon in pairs(oldlist) do if (not list[toon]) then changed=true; break; end end
				if (not changed) then
					for toon in pairs(list) do if (not oldlist[toon]) then changed=true; break; end end
				end
			elseif (guild==GetGuildInfo("player")) then changed=true;		-- always update own guild
			else valid=nil; end												-- not one of mine by AddGuild
			if (valid and changed) then
				if (merge) then list=DM:CopyTable(oldlist,list); end		-- merge old and new
				db[membersDB][guild]=list;	-- pack and store
			end
		end
	end
	util.UpdatingGuildLists=guild;			-- tell MT guild has been received
	dbc.Info.WhoList[guild]=nil;			-- pull it from the queue
end

local function Prune_SpoolLists(src,dst)
	if (not src) then return nil; end
	for _,v in ipairs(src) do table.insert(dst,v); end
--	local i,e,v=1,0,nil;
--	while(e<1000) do								-- 1000 consequtive empties
--		v=src[i];									-- grab this one
--		if (v) then table.insert(dst,v); e=0;		-- add it if it's there
--		else e=e+1; end								-- count it if it's not
--		i=i+1;										-- keep going
--	end
	return true;
end

function util.MT:Prune(prefix)
	local pDB=dbd[prefix];
	local count=self:Prune_FullEntryCount(pDB)-pDB._global_.Limit;		-- Determine amount of pruning
--if (count<0) then say(prefix,"There are",0-count,"entries available in the database.");
--elseif (count>0) then say(prefix,"There are",count,"entries to remove from the database.");
--else say(prefix,"The database is full with no need to remove anything yet."); end
	if (count<=0) then return; end
	local list=self:Prune_MakeList(pDB,count);				-- List X oldest redundant entries
	-- merge sorted lists
	local remlist={};
	local group="r";
	local abort=nil;
	local unused;
	local i;
	repeat
		unused=0;
		i=DM_PRUNE_FIRST; repeat									-- walk prune priorities
			if (not Prune_SpoolLists(list[group][i],remlist)) then	-- append
				unused=unused+1;
			else
				unused=0;
			end
			if (#remlist>=count) then abort=true; break; end
			if (abort) then break; end
			i=i+1;
		until(unused>100);
		if (abort) then break; end
		Prune_SpoolLists(list[group][DM_PRUNE_NORMAL],remlist);	-- append normal list
		if (#remlist>=count) then break; end
		Prune_SpoolLists(list[group][DM_PRUNE_NEVER],remlist);	-- append never list
		if (#remlist>=count) then break; end
		if (group=="a") then break; end							-- both groups done
		group="a";
	until(nil);
	while (#remlist>count) do table.remove(remlist); end		-- delete last until list is short enough
	-- execute list
	for _,e in ipairs(remlist) do
		local thisdb=util:Handle(e.tree);
		if (d_type(thisdb[e.key])=="database") then				-- only empty databases will be added, so they are removable
			util.LockedTime=1;			-- set to not start sync, just something low as it will be wiped immediately anyway
			thisdb[e.key]=nil;			-- set via meta to correctly wipe associated table
			util.LockedTime=nil;		-- back to normal
		end
		thisdb._str_[e.key]=nil;		-- wipe it
	end
end

-- "Prune_MakeList" will seek to add everything that is not a populated
-- database or a link to another database.
-- args are (db,count) in initial call. For subsequent calls,
-- args are (db,count,list).
function util.MT:Prune_MakeList(db,count,list)
	if (not list) then list={ r={}, a={}, }; end			-- init list (r-redundant, a-active)
	db=util:Handle(db);										-- grab it
	if (not db._str_) then return list; end					-- not a database
	if (rawget(db,"_lnk_")) then return list; end			-- it's a link
	local ty;
	for name,entry in pairs(db._str_) do					-- no links
		if (tostring(name):find(hiddenKey,1,true)~=1) then
			ty=entry:match("^%d+:(%a%a):.-:.*$");
			if (ty~="da") then														-- not database
				self:Prune_AddOneTolist(count,list,db,name); DM.MT:Yield(true);		-- add it
			else											-- it's a database
				if (next(db[name]._str_)) then				-- grab real handle and test real storage
					self:Prune_MakeList(db[name],count,list);	-- db with contents so add to list
				else										-- db without contents
					self:Prune_AddOneTolist(count,list,db,name); DM.MT:Yield(true);
				end
			end
		end
	end
	return list;
end

-- entry:
--   tree : the tree
--   key  : the key
--   r    : redundant
--   stamp: stamp
--   p    : prune-mode

--> SORT: if (A<B) then return true; end
-- "redundant" takes precedence over DM_PRUNE_NEVER.
function util.MT:Prune_AddOneTolist(count,list,db,key)
	local s,r=db._str_[key]:match("^(%d+):%a%a:(.-):.*$");
	r=r:find("r",1,true);										-- get setting
	if (db._prn_==DM_PRUNE_NEVER and not r) then return; end	-- non-redundant non-removable, keep it no matter what
	r=(r and "r") or "a";										-- get sub-table name
	list[r][db._prn_]=list[r][db._prn_] or {};					-- create sort priority
	table.insert(list[r][db._prn_],{ tree=db._tree_, key=key, stamp=tonumber(s) });
	table.sort(list[r][db._prn_],function (a,b) return (a.stamp<b.stamp); end );	-- sort by stamp
end

--	DM_PRUNE_NEVER=0,
--	DM_PRUNE_NORMAL=1,
--	DM_PRUNE_FIRST=10,		-- 11 second, 12 third, etc

function util.MT:Prune_FullEntryCount(db)
	local count=0;
	for _,pDB in d_pairs(db,true) do
		count=count+1;
		if (d_type(pDB)=="database") then
			count=count+self:Prune_FullEntryCount(pDB);
			DM.MT:Yield(true);		-- Immediate yield for each database
		end
	end
	return count;
end


--
-- deprecated by adding channel loopback
--
-- dummy event for channels that are slow to get going
--function wq:EMPTY(inter,prefix,tree,key)
--say("From",(inter and inter.ID) or "no ID","or",prefix,"Queue: Empty kickstart");
--	dbd[prefix]=dbd[prefix]-1;	-- Remove directly
--end

-- transmit stuff
function wq:send(inter,prefix,tree,key)
--say("From",(inter and inter.ID) or "no ID","or",prefix,"Queue: Send",util:GetLuaName(tree,key));
	util:SendValue(inter,prefix,tree,key);
	dbd[prefix]=dbd[prefix]-1;
end

-- poll stuff
function wq:poll(inter,prefix,tree,key,ver)
--say("From",(inter and inter.ID) or "no ID","or",prefix,"Queue: Poll",util:GetLuaName(tree,key));
	if (tree=="databaselist") then		-- Poll database list
		util:InterRequest(inter,prefix,ver,nil,{ dblist=true, key=key },nil);	-- set databaselist type
		dbd[prefix]=dbd[prefix]-1;
		return;
	end

	-- if "key" is nil, a database will be polled
	util:InterRequest(inter,prefix,ver,nil,{ tree=tree, key=key },nil);
	dbd[prefix]=dbd[prefix]-1;
end

function dbc:Heartbeat(elapsed)
	dbc.LastPrefixIterator=next(dbd,dbc.LastPrefixIterator);
	if (dbc.LastPrefixIterator) then dbc:Heartbeat_(dbc.LastPrefixIterator,time()); end
end

function dbc:Heartbeat_(prefix,now)
	if (not DM.Talk or not DM.Talk[prefix]) then return; end
	if (not dbc.Addon[prefix]) then return; end

	-- Get next queue item
	local entry=dbd[prefix]("event_queue_peek"); if (not entry or entry.busy) then return; end			-- peek next event
	if (not entry.offset) then entry.offset=0; end						-- extra time offset/hold check

	-- run window
	local delay=0;	-- neg is 2s, may need some consideration here, maybe not as inter should buffer and hold
	if (entry.offset>0) then delay=7; end
	if (now-DM.Talk[prefix]<math.random()*10+delay+entry.offset) then return; end
	entry.busy=true;									-- prevent re-entry for this prefix

	-- event dispatcher
	if (wq[entry.method]) then
		wq[entry.method](wq,entry.inter,prefix,entry.tree,entry.key,entry.instamp);
	else
		dbd[prefix]=dbd[prefix]-1;	-- Remove unknown entry
	end
end

--		util:SendEvent(inter.ID,"DMEVENT_DATABASE_SYNC",goal,meta.key,newdata);
function util:SendEvent(prefix,event,handle,key,newdata)
	if (not dbc.Addon[prefix] or not dbc.Addon[prefix].EventFrame) then return; end
	local handler=dbc.Addon[prefix].EventFrame:GetScript("OnEvent");
	if (not handler) then return; end
--	EventFrame:OnEvent(event,database,key,newdata)
	handler(dbc.Addon[prefix].EventFrame,event,handle,key,newdata);
end

-- accepts both databases and trees
function util:Name(db)
	local tree=db;
	if (d_type(db)=="database") then tree=db._tree_; end
	tree={strsplit(dbc.Separator,tree)};
	for num,entry in ipairs(tree) do
		local loc=entry:sub(2); if (entry:sub(1,1)=="n") then loc=tonumber(loc); end
		tree[num]=loc;
	end
--	tree[1]=self:GetRoot(db);
	return tree;
end

function util:GetLuaName(tree,key)
	if (not key and d_type(tree)=="table") then tree,key=tree.tree,tree.key; end
--print("GetLoc in:",tree:gsub(dbc.Separator,"%^"));
	if (tree=="databaselist") then
		if (key and tostring(key)~="") then return "databaselist ("..key..")"; end
		return "databaselist";
	end
	local tree=util:Name(tree);
	if (key and key~="") then tree[(#tree)+1]=key; end	-- numeric key
	local loc=tree[1];
	for i,value in ipairs(tree) do
--print("Name",i,": ",value);
		if (i>1) then
			if (d_type(value)=="number") then loc=loc.."["..value.."]";
			elseif (not value:find("[%s%-%+]")) then loc=loc.."."..value;
			else loc=loc.."[\""..value.."\"]"; end
		end
	end
	if (not key) then loc=loc.." (db)"; elseif (key~="") then loc=loc.." (key)"; end
	return loc;
end

-- dbd path handler
DM.DatabaseRootMetaTable={
	__index = function (t,k)
		t,k=getDBKEY(k);
		if (t and k) then return t[k]; end
	end,
	__newindex = function (t,k,v)
		t,k=getDBKEY(k); if (not t) then return; end
		if (t==dbd) then
			if (not t[k]) then dbc:Register(k,v); end		-- dbPath["//NEWPREFIX"]=database-or-table
		else t[k]=v; end
	end,
}
setmetatable(dbPath,DM.DatabaseRootMetaTable);				-- set path handler

-- database filesystem
-- "dbtypes" determines type from raw database data without meta
function dbtypes(ref,key,link,human)
--	local entry=ref._str_[key]:match("^%d+:(%a%a):[%+%-].*$");
	local entry=ref._str_[key]:match("^%d+:(%a%a):.-:.*$");
	if (entry=="nu") then return "number";
	elseif (entry=="bo") then return "boolean";
	elseif (entry=="ni") then return "nil";
	elseif (entry=="ta") then return "table";
	elseif (entry=="st") then return "string";
	elseif (entry=="da") then
		if (link and rawget(ref[key],"_lnk_")) then return "symlink"; end
		return (human and "folder") or "database";
	end
	return "other";
end

local function nilhuman(c,text)
	return (c._.human and {text}) or nil;
end

local cmd={ _={}, manpage={} };
function cmd.attrib(c,opt)
	local pa,sw=cmd._.args(opt); if (not pa[1]) then return nil; end
	local handle,key=c._.file(c,pa[1]);							-- get file
	if (not handle or not key or not handle._str_[key]) then return nilhuman(c,key.." not found."); end				-- no such file
	local a,meta,b,update;
	a,meta,b=handle._str_[key]:match("^(%d+:%a%a:)(.-)(:.*)$");
	ti=tonumber(ti);
	-- all
	if (sw.r and not meta:find("r",1,true)) then meta=meta.."r"; update=true; end		-- only add
	if (sw.p) then
		local i,v=sw.p:match("^(%d+),(%d+)$"); i=tonumber(i); v=tonumber(v);
		if (not i or not v or i<1 or i>100) then
			return (c._.human and {"Error in switch: p"}) or nil;
		end
		meta=api.MetaIndex(handle,key,"p",i,v,true);					-- 'true' no sync or storage - only removes current setting and inserts new
		update=true;
	end
	--database
	if (ty=="da") then
		if (sw.e) then handle[key]:DM_EndPoint(value,true); end
		if (sw.E) then handle[key]:DM_EndPoint(value,nil); end
	elseif (sw.e or sw.E) then
		return (c._.human and {"Folders only: e/E"}) or nil;
	end
	if (update) then
		handle._str_[key]=a..meta..b;
		cmd._.sync(handle,key);
	end
	if (c._.human) then return {"Done."}; end
	return dbtypes(handle,key),ti,re;
end
-- cd ..				-- move one up
-- cd <directory>		-- move one down
-- cd ../<directory>	-- move up and down again
-- cd //<absolute>		-- move to absolute position with database specified
-- cd /<absolute>		-- move to absolute position in this database
function cmd.cd(c,loc)
	loc=cmd._.buildpath(c,loc); if (not loc) then return nil; end
	if (not getDB(loc)) then return nil; end
	c._.cur=loc; c._.cur_=getDB(c._.cur);				-- save it
	return (c._.human and {c._.cur}) or true;
end
function cmd.chdir(...) return cmd.cd(...); end
function cmd.copy(c,opt,verb)
	verb=verb or "copied";
	local pa,sw=cmd._.args(opt); if (not pa[1] or not pa[2]) then return nil; end
	local h1,k1=c._.file(c,pa[1]); if (not h1 or not k1 or not h1._str_[k1]) then return nilhuman(c,k1.." not found."); end	-- no such file
	local h2,k2=c._.path(c,pa[2]);
	if (h2) then k2=k1;													-- path exists
	else
		h2,k2=c._.file(c,pa[2]);
		if (not h2) then return nilhuman(c,"Destination folder not found."); end		-- no such path
		if (not k2 or h2._str_[k2]) then return nilhuman(c,k2.." not found."); end		-- destination exists
	end
	if (dbtypes(h1,k1,sw.L)=="symlink") then
		h2:DM_SetLink(k2,util:Handle(h1[k1]));							-- create same link
		if (not c._.human) then return true; end
		return {"Symlink "..k1.." "..verb..", pointing at "..util:Handle(h1[k1]).Path};
	end
	cmd._.sync(h1,k1);
	return (c._.human and {k1.." "..verb.."."}) or true;
end
function cmd.del(c,opt)
	local pa,sw=cmd._.args(opt); if (not pa[1]) then return nil; end
	local handle,key;
	local ret={};
	for _,file in ipairs(pa) do
		handle,key=c._.file(c,file);							-- get file
		if (not handle or not key or not handle._str_[key]) then return nilhuman(c,file.." not found."); end	-- no such file
	end
	if (next(ret)) then return (c._.human and ret) or nil; end
	for _,file in ipairs(pa) do handle,key=c._.file(c,file); handle[key]=nil; table.insert(ret,file.." deleted."); end		-- delete through meta
	return (c._.human and ret) or true;
end
function cmd.deltree(c,opt) return c:rmdir("-s "..opt); end
function cmd.dir(c,opt)
	opt=opt or "";
	if (strtrim(opt)=="//") then									-- list databases
		if (c._.human) then return {"No database specified."}; end
		local list={}; for k in pairs(dbd) do table.insert(list,k); end return list;
	end
	local pa,sw=cmd._.args(opt);
	local pos,da,bo,ta,st,nu,ot,sortN,sortD,reverse=c._.cur,true,true,true,true,true,true,nil,nil,nil;
	if (next(pa)) then pos=pa[1]; end
	if (sw.a) then														-- ==> list specific attributed entries
		if (sw.a:len()>0) then da,bo,ta,st,nu=nil,nil,nil,nil,nil; end	-- there's something, so init
		while(sw.a:len()>0) do
			if (sw.a:sub(1,1)=="f") then da=true;					-- folder
			elseif (sw.a:sub(1,1)=="b") then bo=true;				-- boolean
			elseif (sw.a:sub(1,1)=="t") then ta=true;				-- table
			elseif (sw.a:sub(1,1)=="s") then st=true;				-- string
			elseif (sw.a:sub(1,1)=="n") then nu=true;				-- number
			elseif (sw.a:sub(1,1)=="o") then ot=true;				-- other
			end
			sw.a=sw.a:sub(2);
		end
	end
	if (sw.o) then													-- ==> sorting
		if (sw.o:sub(1,1)=="-") then reverse=true; end				-- reverse sorting
		if (sw.o:sub(-1,-1)=="n") then sortN=true;					-- sort by name
		elseif (sw.o:sub(-1,-1)=="d") then sortD=true; end			-- sort by date/time
	end
	local list,use,pos_={},nil,c._.path(c,pos); if (not pos_) then return nil; end
	for k,v in d_pairs(pos_) do
		use=nil; v=dbtypes(pos_,k);
		if (da and v=="database") then use=true;
		elseif (bo and v=="boolean") then use=true;
		elseif (ta and v=="table") then use=true;
		elseif (st and v=="string") then use=true;
		elseif (nu and v=="number") then use=true;
		elseif (ot and v=="other") then use=true; end
		if (use) then table.insert(list,k); end
	end
	if (sortN) then table.sort(list,function(a,b) if (reverse) then return b<a; end return a<b; end);
	elseif (sortD) then table.sort(list,function(a,b)
--				local timA,timB=tonumber(pos_[a]:match("^(%d+):a%a:[%+%-].*$")),tonumber(pos_[b]:match("^(%d+):a%a:[%+%-].*$"));
				local timA,timB=tonumber(pos_[a]:match("^(%d+):a%a:.-:.*$")),tonumber(pos_[b]:match("^(%d+):a%a:.-:.*$"));
				if (reverse) then return timB<timA; end return timA<timB;
			end);
	end
	if (not c._.human) then return list; end
	-- build output
	local hlist={};
	local long=10;
	for _,v in ipairs(list) do
		if (tostring(v):len()>long) then long=v:len(); end
	end
	table.insert(hlist,c._.cur);
	if (not next(list)) then
		table.insert(hlist,"<empty>");
	else
		for _,v in ipairs(list) do
			local ty=dbtypes(pos_,v,true,true);							-- human
			local v_=tostring(v);
			if (v_:len()>long) then v_=v_:sub(1,long-3).."..."; end
			local ti,me=pos_._str_[v]:match("^(%d+):%a%a:(.-):.*$");
			table.insert(hlist,v_..string.rep(" ",(long+1)-v_:len())..ty..string.rep(" ",10-ty:len())..me.."  "..date("%c",tonumber(ti)));
		end
	end
	return hlist;
end
function cmd.erase(...) return cmd.del(...); end
function cmd.help(c,subject)
	if (not c._.human) then return; end
	subject=subject or "";
	if (cmd.manpage[subject:lower()]) then return cmd.manpage[subject:lower()](); end
	return cmd.manpage.generalhelp();
end
function cmd.label(c)
	local tree=util:Name(c._.cur_);
	return (c._.human and {tree[1]}) or true;
end
function cmd.load(c,opt)
	local pa,sw=cmd._.args(opt); if (not pa[1]) then return nil; end
	local k,v=c._.file(c,pa[1]); if (not k or not v or not k._str_[v]) then return nilhuman(c,v.." not found."); end	-- no such thing
	if (dbtypes(k,v)~="database") then									-- can't load databases (well... you can, but we won't)
		wipe(c._.lastload);
		if (sw.p) then c._.lastload.data=DM.Code:Unprotect(k[v],sw.p);	-- load and apply password
		else c._.lastload.data=k[v]; end								-- normal load
		c._.lastload.db=k; c._.lastload.key=v;							-- bookkeeping
		return (c._.human and {v.." loaded."}) or c._.lastload.data;
	end
end
function cmd.md(...) return cmd.mkdir(...); end
function cmd.mkdir(c,opt)
	local pa,sw=cmd._.args(opt); if (not pa[1]) then return nil; end
	local k,v=c._.file(c,pa[1]);
	if (not k or not v) then return nilhuman(c,"Destination folder not found."); end	-- first argument wrong
	if (k._str_[v]) then return nilhuman(c,v.." already exists."); end	-- already exists
	k[v]={};
	return (c._.human and {"Done."}) or true;
end
function cmd.mklink(c,opt)
	local pa,sw=cmd._.args(opt); if (not pa[1] or not pa[2]) then return nil; end	-- need two path arguments
	local k1,v1=c._.file(c,pa[1]);
	if (not k1) then return nilhuman(c,"Destination folder not found."); end	-- first argument wrong
	if (not v1 or k1._str_[v1]) then return nilhuman(c,v.." already exists."); end	-- already exists
	local k2=c._.path(c,pa[2]); if (not k2) then return nilhuman(c,"Target folder not found."); end
	k1:DM_SetLink(v1,k2);
	return (c._.human and {"Symlink "..v1.." created."}) or true;
end
function cmd.move(c,opt,verb)
	verb=verb or "moved";
	local pa,sw=cmd._.args(opt); if (not pa[1] or not not pa[2]) then return nil; end		-- error
	local ret=c:copy(pa[1].." "..pa[2],nil,verb);						-- copy
	if (ret) then
		if (not c:del(pa[1])) then return nil; end						-- delete source
		return (c._.human and ret) or true;
	end
end
function cmd.prompt(c,opt)
	return (c._.human and {c._.cur}) or c._.cur;
end
function cmd.rd(...) return cmd.rmdir(...); end
function cmd.ren(...) return cmd.rename(...); end
function cmd.rename(c,opt) return cmd.move(c,opt,"renamed"); end
function cmd.rmdir(c,opt)
	local pa,sw=cmd._.args(opt); if (not pa[1]) then return nil; end	-- error
	local h,k=c._.file(c,pa[1]); if (not h or not k or not h._str_[k]) then return nil; end		-- wrong
	if (not h or not k or not h._str_[k]) then return nilhuman(c,k.." not found."); end
	local noremove=nil;
	if (dbtypes(h,k,true)=="database") then
		if (sw.s) then h[k]=nil;										-- deltree
		else
			for key,val in d_pairs(h[k]) do								-- walk it
				if (d_type(val)~="database") then h[key]=nil;			-- delete anything but databases
				else noremove=true; end									-- there are nested databases
			end
			if (not noremove) then h[k]=nil; end
		end
	end
	return (c._.human and {k..((noremove and " deleted.") or " cleared.")}) or true;
end
function cmd.set(c,opt)
	opt=strtrim(opt);
	local ret={};
	for var,val in opt:gmatch("(%S+)=(%S+)") do
		if (tonumber(val)) then val=tonumber(val); end
		c._.cur_:DM_Set(var,val);
		table.insert(ret,var.." set to "..val.." as "..type(val)..".");
	end
	return (c._.human and ret) or true;
end
function cmd.type(...) return cmd.view(...); end
function cmd.view_(list,t,indent)
	indent=indent or "&nbsp; &nbsp; ";
	for k,v in d_pairs(t) do
		table.insert(list,indent..tostring(k).."="..((type(v)=="table" and "") or tostring(v)).."  ("..type(v)..")");
		if (type(v)=="table") then cmd.view_(list,t[k],indent.."&nbsp; &nbsp; "); end
	end
	return list;
end
function cmd.view(c,opt)
	if (not c._.human) then return nil; end
	local pa,sw=cmd._.args(opt); if (not pa[1]) then return nil; end		-- error
	local h,k=c._.file(c,pa[1]);
	if (not h or not k or not h._str_[k]) then return nilhuman(c,k.." not found."); end
	if (dbtypes(h,k)=="database") then
		return {"You can not view folders. Use 'cd' to change to that folder."};
	elseif (dbtypes(h,k)=="table") then
		return cmd.view_({},h[k]);
	end
	return cmd.view_({},{[k]=h[k]});
end
function cmd.ver(c)
	local text=string.format("%.1f",TV);
	return (c._.human and {text}) or text;
end
-- private
function cmd._.sync(db,key)
	local tmp;
	if (dbtypes(db,key)~="database") then tmp=db[key];					-- not a database, so use directly
	else tmp=DM:CopyTable(db[key]); end									-- make copy, later registering will recreate tree
	util.SyncMeta=db._str_[key]:match("^%d+:%a%a:(.-):.*$");			-- time:type:META:data
	db[key]=tmp;														-- save it
end
function cmd._.path(c,loc)
	loc=cmd._.buildpath(c,loc); if (not loc) then return nil; end
	local k=getDB(loc);
	if (not k) then return nil; end			-- try it as database
	return k;								-- return ref
end
function cmd._.file(c,loc)
	loc=cmd._.buildpath(c,loc); if (not loc) then return nil; end
	local k,v=getDBKEY(loc);
	if (not k) then return nil; end			-- try it as database
	return k,v;								-- return ref and key
end
function cmd._.buildpath(c,loc)
	if (not loc) then return c._.cur; end
	loc=strtrim(loc);
	if (loc~="/" and loc:sub(-1)=="/") then loc=loc:sub(1,-2); end	-- got trailing separator
	local pos,here=c._.cur,nil;
	while (loc:sub(1,2)=="..") do
		pos=pos:match("^(.+)/[^/]+$");					-- remove last
		if (not pos) then return nil; end				-- impossible and/or error in argument
		loc=loc:sub(3);									-- remove argument portion
		if (loc:sub(1,1)=="/") then loc=loc:sub(2); end	-- if more to follow, remove separator
	end
	if (loc:sub(1,2)=="//") then				-- absolute path
		pos=loc;
	elseif (loc:sub(1,1)=="/") then				-- relative to root
		pos=pos.."/";							-- just add one to ensure match does not bug out
		pos=pos:match("^(//[^/]+)/");			-- find root
		if (not pos) then return nil; end		-- bugger
		pos=pos..loc;							-- set new absolute for this database
	else pos=pos.."/"..loc; end					-- relative
	if (pos:sub(-1)=="/") then pos=pos:sub(1,-2); end	-- cut trailing slash
	return pos;									-- return absolute path
end
function cmd._.args(arg)
	arg=strtrim(arg or "");
	arg=" "..arg.." ";
	local switch,path,tmp,rem,sw={},{};
	while(arg:len()>2) do
		rem=0;
		tmp=arg:match("^ \"(.-)\" ");						-- quoted file name at beginning
		if (tmp) then rem=tmp:len()+3; table.insert(path,strtrim(tmp));
		else
			tmp=arg:match("^ '(.-)' ");						-- quoted file name at beginning
			if (tmp) then rem=tmp:len()+3; table.insert(path,strtrim(tmp));
			else
				sw,tmp=arg:match("^ %-(.)(.*) ");			-- anything enclosed by spaces, beginning with minus
				if (tmp) then rem=tmp:len()+3; switch[sw]=strtrim(tmp);
				else
					tmp=arg:match("^ (%S*) ");				-- anything non-space enclosed by spaces, not beginning with minus
					if (tmp) then rem=tmp:len()+1; table.insert(path,strtrim(tmp));
					else return nil; end					-- not understandable
				end
			end
		end
		arg=arg:sub(rem+1);
	end
	return path,switch;
end
-- global
function OpenCMD(path,human)
	if (path:sub(-1)=="/") then path=path:sub(1,-2); end	-- cut trailing slash
	local here=getDB(path);
	if (not here) then return nil; end						-- location does not exist
	local c={};
	addclass(cmd,c);
	c._.env={};
	c._.start=path; c._.start_=here;	-- start location at creation
	c._.cur=path; c._.cur_=here;		-- current active location
	c._.human=human;					-- human input and readable output
	c._.lastload={};					-- nothing loaded yet
	return c;
end


-- man-pages
function cmd.manpage.attrib()
	return {[[
Displays or changes entry and folder attributes.

attrib [-r] [-e] [-E] [-pI,V] [path]entry
 -r    Make entry redundant.
 -e    Folder only: Make folder an end point.
 -E    Folder only: Remove end point setting.
 -pI,V Property assignment, I = index, V = value
]]}
end
function cmd.manpage.cd()
	return {[[
Changes the current folder.

cd path
chdir path
]]}
end
function cmd.manpage.chdir() return cmd.manpage.cd(); end
function cmd.manpage.copy()
	return {[[
Copies an entry or a folder to another location.

copy [-L] source destination
 -L  If the source is a symbolic link, copies the link instead of the folder it points to.
]]}
end
function cmd.manpage.del()
	return {[[
Deletes one or more entries.

erase entry1 entry2 ... entryN
del entry1 entry2 ... entryN
]]}
end
function cmd.manpage.deltree()
	return {[[
Deletes a folder and all its contents.

deltree folder
]]}
end
function cmd.manpage.dir()
	return {[[
Displays a list of entries and folders in a folder.

dir [-a[f][b][t][s][n][o]].."]]"..[[ [-o[-][n][d]].."]]"..[[

 -a  List only selected entry types
     f Folder
     b Boolean
     t Table
     s String
     n Number
     o Other
 -o  List order
     - Reverse order
     n Sort by name
     d Sort by date/time
]]}
end
function cmd.manpage.erase() return cmd.manpage.del(); end
function cmd.manpage.help()
	return {[[
Displays help information for CMD.

help [command]
]]}
end
function cmd.manpage.label()
	return {[[
Displays the current database name.

label
]]}
end
function cmd.manpage.load()
	return {[[
Load an entry into memory.

load [-p<password>] entry
 -p  Specify password for password protected entries.
]]}
end
function cmd.manpage.md() return cmd.manpage.mkdir(); end
function cmd.manpage.mkdir()
	return {[[
Creates a folder.

md folder
mkdir folder
]]}
end
function cmd.manpage.mklink()
	return {[[
Creates a symbolic link.

mklink link folder
]]}
end
function cmd.manpage.move()
	return {[[
Moves an entry from one location to another.

move entry destination
]]}
end
function cmd.manpage.prompt()
	return {[[
Displays the current location.

prompt
]]}
end
function cmd.manpage.rd() return cmd.manpage.rmdir(); end
function cmd.manpage.ren() return cmd.manpage.rename(); end
function cmd.manpage.rename()
	return {[[
Renames an entry or folder.

ren entry1 entry2
rename entry1 entry2
]]}
end
function cmd.manpage.rmdir()
	return {[[
Removes a folder.

rd folder
rmdir folder

If the folder contains other folders it will only delete all contents but folders.
]]}
end
function cmd.manpage.set()
	return {[[
Set a database environment variable.

set variable1=value1 variable2=value2 ... variableN=valueN

NOTE: The variable persists across sessions and will not be shared among other users.
]]}
end
function cmd.manpage.ver()
	return {[[
Displays database library version.

ver
]]}
end
function cmd.manpage.view()
	return {[[
Displays the contents of an entry.

view entry
]]}
end
function cmd.manpage.generalhelp()
	return {[[
For more information on a specific command, type 'help <command-name>'
attrib  Displays or modifies entry attributes.
cd      Changes the current folder.
chdir   Change the current folder.
copy    Copies an entry or a folder to another location.
del     Deletes one or more entries and/or folders.
deltree Removes or empties a directory.
dir     Displays a list of entries and folders in a directory.
erase   Deletes one or more entries and/or folders.
exit    Exit command interpreter.
help    Provides help for CMD.
label   Displays the name of your active database.
load    Loads an entry from the database.
md      Creates a folder.
mkdir   Creates a folder.
mklink  Creates a symbolic link to a folder.
move    Moves an entry or folder to another location.
prompt  Displays current location.
rd      Removes a folder.
ren     Renames an entry or folder.
rename  Renames an entry or folder.
rmdir   Removes a folder.
set     Sets a database environment value.
type    Displays the contents of an entry.
ver     Displays database library version.
view    Displays the contents of an entry.
]]};
end
