--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2014  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]


local TV=3.0;

if (not DuckMod) then DuckMod = {}; end			-- Make it if empty
if (not DuckMod[TV]) then					-- Don't mess with it if it's loaded
	DuckMod[TV]={};
end
local _;

local DUCKNET_HEARTBEAT = 1/5;			-- 200ms ticker (5 fps)

-- 1. WoWnet server + search (one instance only / private for WoWnet / global API)
-- 1a. Addon server (available to instantiate for addons?)
-- 3. Client (available to instantiate for addons)

DuckMod[TV].WoWnet={
	WoWnet={			--> WoWnet server
		Ticker=0,		-- heartbeat timer
		Input={},		-- incoming request handlers
	},
	Server={			--> Addon interface to WoWnet server
		List={},		-- existing entries
	},
	Proxy={				--> Alt toons handling
		Hop={},			-- known proxies --==>> NOTE: METATABLE ADDED
		Pending={},		-- pending queries
	},


	Client={			--> Client Sockets
		List={},		-- existing sockets
		API={},			-- socket api
	},
};
local noreply=DuckMod[TV].NoReply;
local c2=DuckMod[TV].Com2;
local wn=DuckMod[TV].WoWnet;
setmetatable(wn.Proxy.Hop,{ __index=function(t,k) return k; end });		-- return key if key does not have a proxy

local WN_ID="wn";
local WN_CHN="WoWnet";

local function dbg(h,t)
	print(h,":");
	for _i,_d in pairs(t) do print(_i,_d); end
end
local function addclass(class,container)
	for k,v in pairs(class) do container[k]=v; end				-- add class to container
end
local function copytable(t)
	if (not t) then return nil; end
	local new={};
	for i,v in pairs(t) do
		if (type(v)=="table") then new[i]=copytable(v,new[i]); else new[i]=v; end
	end
	return new;
end


--==>>
--==>> WoWnet server
--==>>
function wn:Tick(elapsed)
	wn.WoWnet.Ticker=wn.WoWnet.Ticker-elapsed;
	if (wn.WoWnet.Ticker>=0) then return; end
	wn.WoWnet.Ticker=DUCKNET_HEARTBEAT;							-- slower, inaccurate, lower CPU impact
	local now=time();
	for player,pdata in pairs(wn.Proxy.Pending) do		-- pending queries
		if (pdata.queue==1) then								-- nothing done yet
			if (now-pdata.started>2) then						-- it's been pending for over two seconds
				if (not noreply[player]) then					-- no error message detected
					wn.Proxy.Pending[player]=nil;		-- assume everything worked out fine
					return;
				end
				wn:FindToon(player);							-- search for it in case of proxy (available through an alt)
				pdata.queue=2;
			end
		elseif (pdata.queue==2) then							-- a toon search was started
			if (now-pdata.started>4) then						-- it's been pending for over two more seconds
				if (wn.Proxy.Hop[player]==player) then	-- no proxy found
					-- give "toon not found" message
				else											-- a proxy exists, so reattempt communication
					local buf={};
					for i=1,pdata.n do buf[i]=pdata[i]; end		-- do all entries - even if nil
					wn.WoWnet:privateOut(player,unpack(buf,1,pdata.n));	-- spew with predefined table length to preserve nil
				end
				wn.Proxy.Pending[player]=nil;			-- can't do anything more
				return;
			end
		end
	end
end

function wn:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-WN-Messager"..TV,UIParent);
		if (not self.TheFrame) then DM:Chat("Could not create DuckMod WN UI frame",1); return; end
		self.TheFrame:SetScript("OnUpdate",self.Tick);
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-WNWM-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then DM:Chat("Could not create DuckMod WN WM frame",1); return; end
		self.TheFrameWorldMap:SetScript("OnUpdate",self.Tick);
	end
	self.WoWnet:RunWN();
end

-- this function effectively starts the server with public communication
function wn.WoWnet:RunWN()
	self.address=UnitName("player"):lower()..string.format("%X",tostring(math.random(0x7FFF)));
	self.protocol="wownet";
	self.Reply={};
	wn.CallbackProtocol={};
	wn.CallbackSearch={};
	-- WHISPER input
	self.wIn=function(_,data,name,_,sender,_)
		local _adr,_pro,_too,_pla,_fil=wn:ResolveLink(data.link);		-- resolve requested goal
		if (name=="GO") then											-- it's trying to go somewhere
			_fil=_fil or "index";										-- answer with "index" if none specified
			for k,v in ipairs(wn.WoWnet.Server.List) do					-- so check what we've got
				if (_adr==v.address and _pro==v.protocol and _too==v.toon and _fil==v.file) then	-- I got this one
					local result=nil;
					if (type(v.contents)=="function") then result=v.contents(data.i,_fil,_too,_pro,_adr);	-- note the reverse order
					else result=v.contents; end
					if (result) then
						self:privateOut(sender,{link=wn:FormatLink(v.address,v.protocol,v.toon,v.file,"-"),d=result},"contents");
					end
				end
			end
			return;
		end
		if (name=="contents") then										-- some addon specific data/contents incoming
			if (not _pro or wn.CallbackProtocol[_pro]) then return; end	-- protocol or callback not found for data contents
			wn.CallbackProtocol[_pro](_pro,wn:FormatLink(_add,_pro,_too,_fil,nil),data.d,sender);	-- notify protocol-capable addon
			return;
		end
		if (name=="proxy") then							-- a proxy entry
			if (data.proxyFor) then						-- valid
				if (data.proxyFor==sender) then wn.Proxy.Hop[data.proxyFor]=nil;	-- no proxy
				else wn.Proxy.Hop[data.proxyFor]=sender; end						-- set proxy
			end
			return;
		end
		if (name~="result") then return; end			-- not search
		local p;
		for k,v in pairs(self.Reply) do
			if (v[data.source]) then
				v[data.source]=v[data.source] or {}; p=v[data.source];
				p[sender]=p[sender] or {}; p=p[sender];
				p[_adr]=p[_adr] or {}; p=p[_adr];
				p[_pro]=p[_pro] or {}; wipe(p[_pro]);
				p[_pro].scope=data.scope;
				p[_pro].description=data.description;
				if (wn.CallbackSearch[k]) then wn.CallbackSearch[k](data.source,copytable(v[data.source])); end
			end
		end
	end

	-- create listening interface for PUBLIC INPUT
	self.public=c2:CreateInterface(WN_ID,self.ServerRX,nil,nil,nil,"CHANNEL");	-- input from client public comm
	self.public:SetChannel("CHANNEL",WN_CHN);
	-- create sending interface for PRIVATE REPLIES to searches
	self.private=c2:CreateInterface(WN_ID,self.wIn,nil,nil,nil,"WHISPER");
end

function wn.WoWnet:privateOut(toon,...)
	local player=wn.Proxy.Hop[toon];
	self.private:SetChannel("WHISPER",player);
	if (player==toon) then
		wn.Proxy.Pending[player]=wn.Proxy.Pending[player] or {}; wipe(wn.Proxy.Pending[player]);	-- reuse
		local p=wn.Proxy.Pending[player];
		p.started=time();
		p.queue=1;				-- nothing done yet
		p.toon=toon;
		p.n=select("#",...);
		for i=1,p.n do p[i]=select(i,...); end
	end
	self.private:Send(...);
end

function wn:FindToon(toon)
	if (type(prot)~="string") then return; end
	self.WoWnet.public:Send(toon,"findtoon");
end

function wn:ClearResult(addon,term)
	self.WoWnet.Reply[addon]=self.WoWnet.Reply[addon] or {};
	self.WoWnet.Reply[addon][term]=self.WoWnet.Reply[addon][term] or {};
	wipe(self.WoWnet.Reply[addon][term]);
end


-- ServerRX dispatch
function wn.WoWnet.Input.findprot(protocol,sender)
	for _,v in pairs(wn.WoWnet.Server.List) do
		if (v.protocol:find(protocol,1,true)) then
			v.private:Send({source=protocol,scope="protocol",description=v.description,link=wn:FormatLink(v.address,v.protocol,v.toon,v.file,"-")},"result",sender);
		end
	end
end

-- ServerRX dispatch
function wn.WoWnet.Input.findadd(address,sender)
	for _,v in pairs(wn.WoWnet.Server.List) do
		if (v.address:find(address,1,true)) then
			v.private:Send({source=address,scope="address",description=v.description,link=wn:FormatLink(v.address,v.protocol,v.toon,v.file,"-")},"result",sender);
		end
	end
end

-- ServerRX dispatch
function wn.WoWnet.Input.findtoon(toon,sender)
	if (UnitName("player"):lower()==toon:lower()) then
		v.private:Send({source=toon,scope="toon",description=v.description,proxyFor=UnitName("player")},"proxy",sender);
		return;
	end
	for _,v in pairs(wn.WoWnet.Server.List) do
		if (v.toon:lower()==toon:lower()) then
			v.private:Send({source=toon,scope="toon",description=v.description,proxyFor=v.toon},"proxy",sender);
			return;
		end
	end
end

-- ServerRX dispatch
function wn.WoWnet.Input.search(search,sender)
	for _,v in pairs(wn.WoWnet.Server.List) do
		if ((v.address:find(search,1,true) or v.description:lower():find(search,1,true))) then
			v.private:Send({source=search,scope="search",description=v.description,link=wn:FormatLink(v.address,v.protocol,v.toon,v.file,"-")},"result",sender);
		end
	end
end

-- ServerRX entry-point
function wn.WoWnet.ServerRX(inter,data,name,_,sender,_)			-- inter,data,name,metadata,sender,api)
	if (type(data)~="string") then
print("Public data in can't be",type(data));
		return;
	end
print("ServerRX from",sender);
print("Data",name,"-",data);
	data=strtrim(data):lower();
	if (wn.WoWnet.Input[name]) then wn.WoWnet.Input[name](data,sender); end
end


--==>>
--==>> Addon interface to WoWnet server
--==>>
function wn.Server:Add(address,protocol,toon,file,contents,description)
	if (not contents) then return; end
	if (not name or name=="") then name="index"; end
	if (file:sub(1,1)~="/") then location="/"..file; end
	table.insert(wn.WoWnet.Server.List,{address=address,protocol=protocol,toon=toon,file=file,contents=contents,description=description});
end

function wn.Server:Remove(address,protocol)
	for k,v in ipairs(wn.WoWnet.Server.List) do
		if (v.address==address and v.protocol==protocol) then
			table.remove(wn.WoWnet.Server.List,k);							-- unregister
			return;
		end
	end
end


--==>>
--==>> Addon interface to WoWnet
--==>>
function wn:SetProtocol(protocol,callback)
	self.CallbackProtocol[protocol]=callback;
end

function wn:SetSearch(addon,callback)
	self.CallbackSearch[addon]=callback;
end

function wn:FindProtocol(addon,prot)
	if (type(prot)~="string") then return; end
	self.WoWnet:ClearResult(addon,prot);
	self.WoWnet.public:Send(prot,"findprot");
end

function wn:FindAddress(addon,add)
	if (type(add)~="string") then return; end
	self.WoWnet:ClearResult(addon,add);
	self.WoWnet.public:Send(add,"findadd");
end

function wn:Request(addon,search)
	if (type(search)~="string") then return; end
	local add,prot,toon,file=wn:ResolveLink(search);
	if (not add or not prot or not toon) then		-- can't resolve properly so perform search
		self:ClearResult(addon,search);
		self.WoWnet.public:Send(search,"search");
	else											-- resolved, so hail toon
		if (not file) then file="index"; end
		wn.WoWnet:privateOut(toon,{link=wn:FormatLink(add,prot,toon,file,"-"),i=info},"GO");
	end
end


-- complete link widget presentation: "|cff808080|Hwownet:completeaddress|h[visual text]|h|r"
-- full link: wownet:address.protocol@toon/file
-- options: wownet:address[.protocol][@toon][/file]
function wn:FormatLink(address,protocol,toon,file,displaytext)
	if (not address) then return displaytext or ""; end
	local link="wownet:"..address;
	if (protocol) then link=link.."."..protocol; end
	if (toon) then link=link.."@"..toon; end
	if (file) then link=link.."/"..file; end
	if (not displaytext) then displaytext=link; end
	return "|cffA0FFA0|H"..link.."|h["..displaytext.."]|h|r";
end

function wn:ResolveLink(link)
	if (not link) then return nil; end
	link=strtrim(link);
	local _adr,_pro,_too,_pla,_fil;
	local colon=link:find("wownet:",1,true); if (not colon) then return nil; end
	colon=colon+6;					-- find colon
	local pipe=link:find("|",pos,true); if (pipe) then link=link:sub(1,pipe-1); end
	local space=link:find(" ",pos,true); if (space) then link=link:sub(1,space-1); end
	local dot=link:find(".",pos,true);
	local at=link:find("@",pos,true);
	local excl=link:find("!",pos,true);
	local slash=link:find("/",pos,true);
	if (slash) then _fil=link:sub(slash+1); link=link:sub(1,slash-1); end
	if (excl) then _pla=link:sub(excl+1); link=link:sub(1,excl-1); end
	if (at) then  _too=link:sub(at+1); link=link:sub(1,at-1); end
	if (dot) then  _pro=link:sub(dot+1); link=link:sub(1,dot-1); end
	_adr=link:sub(colon+1);
	return _adr,_pro,_too,_pla,_fil;
end
