
local ver = 0.00
Matrix=Matrix or {};
if (not Matrix[ver]) then Matrix[ver]={ api={ Set={} }, gui={}, design={}, }; end
local mx=Matrix[ver];


function mx.copytable(t,new)
	if (not t and not new) then return nil; end
	t=t or {}; new=new or {};
	for k,v in pairs(t) do if (type(v)=="table") then new[k]=mx.copytable(v,new[k]); else new[k]=v; end end
	return new;
end
function mx.addclass(class,container)
	for k,v in pairs(class) do container[k]=v; end				-- add class to container
end
function mx.GetFunctionReferenceByText(f)
	if (f:find("(",1,true)) then f=loc:match("(.+)%(.*%)"); end	-- strip paranthesis and arguments
	f=strtrim(f);
	local p,s;
	local fu={strsplit(":",f)};									-- grab class and member
	local fun={strsplit(".",fu[1])};							-- grab nested table keys
	for k,v in ipairs(fun) do if (k==1) then p=_G[v]; else p=p[v]; end end	-- traverse class or function
	if (fu[2]) then s=p;										-- it's a class, so set self to class reference
		fun={strsplit(".",fu[2])};								-- grab nested class member keys
		for _,v in ipairs(fun) do p=p[v]; end					-- traverse class members
	end
	return p,s;													-- return function reference and self (if any)
end


-- load GUI from provided storage collection of monitors
function mx.Load(mem)
	for _,u in pairs(mem) do									-- walk unit settings in mem
		if (type(u)~="table") then mem={mem}; break; end		-- convert single unit setting to collection
	end
	for _,u in ipairs(mem) do									-- walk unit settings in mem
		local frame=mx.CreateGUI(reg.mode);						-- create frame and links to storage: nested table "storage" in "frame".
		frame:Initialize(u);									-- shove compatible settings
	end
end

function mx.CreateGUI(mode,design)
	local reuse,frame=nil,nil;
	for _,fr in ipairs(mx.gui) do if (not fr:IsVisible()) then frame=fr; reuse=true; end end		-- reuse a hidden monitor
	if (not reuse) then
		frame=CreateFrame("Frame",nil,UIParent);
		frame:EnableMouse(true);
		frame:SetMovable(true);
		frame._UI={};
		frame._UI.Background=frame:CreateTexture(); frame._UI.Background:SetDrawLayer("BACKGROUND");
		frame._UI.Icon=frame:CreateTexture();
		frame._UI.Overlay=frame:CreateTexture(); frame._UI.Overlay:SetDrawLayer("OVERLAY");
		frame._UI.Header=frame:CreateFontString();
		frame._UI.Header:SetFontObject("GameFontNormal");
		frame._UI.Counter=frame:CreateFontString();
		frame._UI.Counter:SetFontObject("GameFontNormal");
		frame._UI.Bar=frame:CreateTexture("BACKGROUND");
		frame.FullWidth=1;
		mx.addclass(mx.api,frame);
	end
	frame._UI.Bar:SetTexture("Interface\\AddOns\\LootCount_DropCount\\white.tga");
	frame._UI.Bar:SetVertexColor(.4,.4,1,.6);		-- r,g,b,a
	frame:SetDesign(design);
	frame.Ticker=0;
	frame.storage=frame.storage or {}; wipe(frame.storage);
	frame.storage.max=0;
	frame:SetMode(mode or "default");
	frame:SetUpdate();						-- no update cb yet
	frame:PositionMonitor();
	frame:SetScript("OnUpdate",mx.api._OnUpdate);
	frame:SetScript("OnMouseDown",mx.api._OnClick);
	frame:SetScript("PreClick",mx.api._PreClick);
	frame:SetScript("OnReceiveDrag",function(frame)
		if (frame.storage.locked) then return; end						-- locked - no change allowed
		frame.storage.dropped={GetCursorInfo()} ClearCursor();			-- grab and clear cursor contents
		frame:Set[frame.storage.dropped[1]](frame.storage.dropped[2],frame.storage.dropped[3],frame.storage.dropped[4]);
		end );
	frame:SetScript("OnDragStart",function(frame,LR)
			if (frame.storage.locked) then return; end
			if (LR=="LeftButton") then
				frame:StartMoving();
				if (not frame.storage.freemove) then frame:TagFrameGroup("glued"); end		-- got glue on edges?
			end
		end);
	frame:SetScript("OnDragStop",function(frame,LR)
			frame:StopMovingOrSizing();
			frame:PositionMonitor(frame.startL+mx.api.offsetL,frame.startB+mx.api.offsetB);
			if (not frame:UntagFrameGroup("glued",true)) then frame:Snap(); end		-- none glued, snap normally
		end);
	frame:RegisterForDrag("LeftButton");	-- moving
	if (not reuse) then table.insert(mx.gui,frame); end
--	frame._UI.Icon:SetTexture("Interface\\Icons\\INV_Misc_QuestionMark");
	if (frame.storage.dropped) then										-- something was dropped on it so do that
		frame:Set[frame.storage.dropped[1]](frame.storage.dropped[2],frame.storage.dropped[3],frame.storage.dropped[4]);
	else
		frame.storage.autoitem=true;
	end
	frame:SetPercent(0);
	return frame;
end

function mx.api.Initialize(frame,mode,update,delay,name,design,left,bottom)
	if (type(mode)=="table") then
		update=mode.update;
		delay=mode.delay;
		left=mode.left;
		bottom=mode.bottom;
		name=mode.name;
		design=mode.design;
		mode=mode.mode;
	end
	frame.storage.name=name;
	frame:SetMode(mode);
	frame:SetDesign(design);
	frame:SetUpdate(delay,update);
	frame:PositionMonitor(left,bottom);							-- if position is absent, a predetermined position will be used
end

--		if (frame.storage.dropped[1]) then						-- there's drop registered
--			frame:Set[frame.storage.dropped[1]](frame.storage.dropped[2],frame.storage.dropped[3],frame.storage.dropped[4]);
--		end


function mx.api.SetDesign(frame,design)
	local d;
	if (not mx.design[design]) then
		d={
			Background=nil,
			Icon="Interface\\Icons\\INV_Misc_QuestionMark",
			Overlay=nil,
		};
	else
		d=mx.copytable(mx.design[design]);
	end
	frame.storage.design=d;
	frame._UI.Background:SetTexture(d.Background or "");
	frame._UI.Icon:SetTexture("Interface\\Icons\\INV_Misc_QuestionMark");
	frame._UI.Background:SetTexture(d.Overlay or "");
end

function mx.api.AddDesign(frame,design,data)
	mx.design[design]=mx.copytable(data);
end

function mx.api.SetUpdate(frame,delay,cb)
	if (not delay) then
		delay=frame.storage.delay;
		cb=frame.storage.update;
	elseif (not cb) then
		cb=frame.storage.update;
	end
	if (not delay or not cb) then return; end
	if (type(cb)=="string") then			-- function name provided
		frame.storage.delay=delay;
		frame.storage.update=cb;
		frame.cbUpdate,frame.cbUpdateSelf=mx.GetFunctionReferenceByText(cb);
		frame.cbDelay=delay;
	else									-- function reference provided - can't save that, but can use for a single session
		frame.storage.delay=nil;
		frame.storage.update=nil;
		frame.cbUpdate=cb;
		frame.cbDelay=delay;
	end
end

function mx.api.SetMode(frame,mode)
	mode=(mode or "default"):lower();
	frame.storage.mode=mode;
	dccs.LastMonitorMode=mode;
--	frame:ClearAllPoints();
	frame._UI.Icon:ClearAllPoints();
	frame._UI.Header:ClearAllPoints();
	frame._UI.Counter:ClearAllPoints();
	frame._UI.Bar:ClearAllPoints();
	if (mode=="compact") then		-- mode: compact
		frame:SetBackdrop(nil);
		frame:SetSize(32,32);
		frame._UI.Icon:SetPoint("TOPLEFT",frame,"TOPLEFT");
		frame._UI.Icon:SetPoint("BOTTOMRIGHT",frame,"BOTTOMRIGHT");
		frame._UI.Counter:SetJustifyH("RIGHT");
		frame._UI.Counter:SetPoint("BOTTOMLEFT",frame._UI.Icon,"BOTTOMLEFT");
		frame._UI.Counter:SetPoint("RIGHT",frame._UI.Icon,"RIGHT");
		frame._UI.Bar:SetPoint("TOPLEFT",frame._UI.Counter,"TOPLEFT");
		frame._UI.Bar:SetPoint("BOTTOM",frame._UI.Counter,"BOTTOM");
		frame._UI.Bar:SetWidth(0);
		frame._UI.Header:Hide();
	else							-- mode: default
		frame:SetBackdrop(nil);
		frame:SetSize(32,32);
		frame._UI.Icon:SetPoint("TOPLEFT",frame,"TOPLEFT");
		frame._UI.Icon:SetPoint("BOTTOMRIGHT",frame,"BOTTOMRIGHT");
		frame._UI.Header:SetPoint("LEFT",frame._UI.Icon,"RIGHT",3,0);
		frame._UI.Header:SetPoint("BOTTOM",frame._UI.Icon,"CENTER");
		frame._UI.Counter:SetJustifyH("LEFT");
		frame._UI.Counter:SetPoint("TOP",frame._UI.Icon,"CENTER");
		frame._UI.Counter:SetPoint("LEFT",frame._UI.Header,"LEFT");
		frame._UI.Bar:SetPoint("TOPLEFT",frame._UI.Counter,"TOPLEFT");
		frame._UI.Bar:SetPoint("BOTTOM",frame._UI.Counter,"BOTTOM");
		frame._UI.Bar:SetWidth(0);
		frame._UI.Header:Show();
	end
	frame:Show();
end



function mx.api.PositionMonitor(frame,L,B)
	frame.storage.left=L or (UIParent:GetWidth()/2);
	frame.storage.bottom=B or (UIParent:GetHeight()*0.75);
	frame:ClearAllPoints();
	frame:SetPoint("BOTTOMLEFT",UIParent,"BOTTOMLEFT",frame.storage.left or (UIParent:GetWidth()/2),frame.storage.bottom or (UIParent:GetHeight()*0.75));
end

-- tag all monitors glued to the supplied monitor - must use "glue edge" setting, snapped position is not enought
function mx.api.TagFrameGroup(frame,tag,verticalOnly)
	local fL,fB=frame.storage.left,frame.storage.bottom;
	local frL,frB;
	local grouped=nil;
	local slack=.5;			-- WoW "pixels" operates with fractions
	for _,fr in ipairs(mx.gui) do
		if (fr[tag]) then grouped=true; end
		if (fr~=frame and fr:IsVisible() and not fr.storage.freemove and not fr[tag]) then
			frL,frB=fr.storage.left,fr.storage.bottom;
			if (frL>fL-slack and frL<fL+slack) then							-- same horisontal
				if (frB>(fB+frame:GetHeight())-slack and frB<(fB+frame:GetHeight())+slack) then fr[tag]=true; grouped=true; fr:TagFrameGroup(tag,verticalOnly); end
				if (frB>(fB-frame:GetHeight())-slack and frB<(fB-frame:GetHeight())+slack) then fr[tag]=true; grouped=true; fr:TagFrameGroup(tag,verticalOnly); end
			end
			if (not verticalOnly and frB>fB-slack and frB<fB+slack) then	-- same vertical
				if (frL>(fL+frame:GetWidth())-slack and frL<(fL+frame:GetWidth())+slack) then fr[tag]=true; grouped=true; fr:TagFrameGroup(tag,verticalOnly); end
				if (frL>(fL-frame:GetWidth())-slack and frL<(fL-frame:GetWidth())+slack) then fr[tag]=true; grouped=true; fr:TagFrameGroup(tag,verticalOnly); end
			end
		end
	end
	frame[tag]=grouped;			-- tag yourself if there are others
	return grouped;
end

-- untag all monitors and re-evaluate position for all grouped monitors
function mx.api.UntagFrameGroup(frame,tag,reposition)
	local found=nil;
	for _,fr in ipairs(mx.gui) do
		if (fr~=frame and fr:IsVisible() and fr[tag]) then
			if (reposition) then
				fr:PositionMonitor(frame.storage.left-fr.osL,frame.storage.bottom-fr.osB);	-- offset to dragged frame
			end
			found=true;
		end
		fr.startL=nil; fr.startB=nil; fr[tag]=nil; fr.osL=nil; fr.osB=nil;
	end
	return found;
end

-- Sets a unit to some contents where parameters match what is returned from GetCursorInfo()
-- battlepet
-- companion
-- equipmentset
-- flyout
-- guildbankmoney
function mx.api.Set.item(frame,itemID,link)
	if (not itemID and not link) then
		itemID,link=frame.storage.dropped[2],frame.storage.dropped[3];	-- nothing provided so use what's there
	end
	if (frame.FollowItem==link) then frame.Ticker=1; return; end		-- same item set again, just update
	frame.storage.autoitem=nil;
	frame.FollowItem=link;
	frame._UI.Icon:SetTexture((link and GetItemIcon(tonumber(link:match("item:(%d+)")))) or "Interface\\Icons\\INV_Misc_QuestionMark");
	if (not link) then
		frame._UI.Header:SetText("");
		frame._UI.Counter:SetText("");
		return;
	end
	if (not DropCount.Tracker.MaxSessionValue[link]) then DropCount.Tracker.MaxSessionValue[link]=0; end
	if (not GetItemInfo(link)) then								-- unknown item
		frame.ItemNotInCache=true;
--		DropCount.Cache:AddItem(item);							-- queue for cache
		return;
	end
	frame.ItemNotInCache=nil;
	local name=GetItemInfo(link);
	frame._UI.Header:SetText(name);
	frame:SetPercent((DropCount:TimedQueueRatio(getid(frame.FollowItem),true) or 0)*100);		-- item,noyield
	frame:Regroup();

	frame.Ticker=0;
end
-- macro
-- merchant
-- money
-- petaction
function mx.api.Set.spell(frame,index,book,spellID)		-- book index, "spell"/"pet" book, spellID
end

function mx.api.Snap(frame)
	local snap,snapL,snapT=15,nil,nil;
	frame.storage.left=frame:GetLeft(); frame.storage.bottom=frame:GetBottom();
	-- left
	local span,tmp=frame:GetWidth();
	for _,reg in ipairs(dccs.Monitor) do if (reg~=frame.storage) then		-- don't snap to yourself
		tmp=frame.storage.left-reg.left;
		if (tmp<span+snap and tmp>span-snap) then
			frame:PositionMonitor(reg.left+span,frame.storage.bottom);
			snapL=true;
		end
		tmp=reg.left-frame.storage.left;
		if (tmp<span+snap and tmp>span-snap) then
			frame:PositionMonitor(reg.left-span,frame.storage.bottom);
			snapL=true;
		end
		if (abs(frame.storage.left-reg.left)<snap) then
			frame:PositionMonitor(reg.left,frame.storage.bottom);
			snapL=true;
		end
	end end
	-- bottom
	span=frame:GetHeight();
	for _,reg in ipairs(dccs.Monitor) do if (reg~=frame.storage) then		-- don't snap to yourself
		tmp=frame.storage.bottom-reg.bottom;
		if (tmp<span+snap and tmp>span-snap) then
			frame:PositionMonitor(frame.storage.left,reg.bottom+span);
			snapT=true;
		end
		tmp=reg.bottom-frame.storage.bottom;
		if (tmp<span+snap and tmp>span-snap) then
			frame:PositionMonitor(frame.storage.left,reg.bottom-span);
			snapT=true;
		end
		if (abs(frame.storage.bottom-reg.bottom)<snap) then
			frame:PositionMonitor(frame.storage.left,reg.bottom);
			snapT=true;
		end
	end end
	-- edge
	if (not snapL) then
		if (frame.storage.left<(snap/2)) then
			frame:PositionMonitor(0,frame.storage.bottom);
			snapL=true;
		elseif (frame.storage.left+frame:GetWidth()>frame:GetParent():GetRight()-(snap/2)) then
			frame:PositionMonitor(frame:GetParent():GetRight()-frame:GetWidth(),frame.storage.bottom);
			snapL=true;
		end
	end
	if (not snapT) then
		if (frame.storage.bottom<(snap/2)) then
			frame:PositionMonitor(frame.storage.left,0);
			snapT=true;
		elseif (frame.storage.bottom+frame:GetHeight()>frame:GetParent():GetTop()-(snap/2)) then
			frame:PositionMonitor(frame.storage.left,frame:GetParent():GetTop()-frame:GetHeight());
			snapT=true;
		end
	end
	-- parked
	for _,reg in ipairs(dccs.Monitor) do if (reg~=frame.storage) then		-- don't snap to yourself
		if (frame.storage.left==reg.left and frame.storage.bottom==reg.bottom) then
			frame:PositionMonitor();
			snapL=true; snapT=true;
		end
	end end

	frame:Regroup();

	return snapL or snapT;
end

-- equalize all bar sizes and modes within all groups
function mx.api.Regroup(master)
	-- set all modes
	for _,mon in ipairs(mx.gui) do											-- do all monitors
		if (mon:IsVisible() and mon:TagFrameGroup("group")) then			-- locate this monitor's group, if any
			-- grab settings from the group
			for _,fr in ipairs(mx.gui) do if (fr.group and fr~=m) then		-- grab anyone else from the group
				if (master.group) then fr=master; end					-- we're doing the master's group
				-- override all
				for _,frm in ipairs(mx.gui) do if (frm.group) then			-- walk all tagged monitors
					frm.storage.max=fr.storage.max;						-- copy max bar setting
					frm:SetMode(fr.storage.mode);						-- copy display mode
				end end
				break;
			end end
			mon:UntagFrameGroup("group");								-- remove tag
		end
	end

	-- set all sizes
	for _,mon in ipairs(mx.gui) do											-- do all monitors
		if (mon:IsVisible() and mon:TagFrameGroup("group",true)) then		-- locate this monitor's vertical group members, if any
			local width,bw=0,0;
			-- find biggest
			for _,fr in ipairs(mx.gui) do if (fr.group) then				-- walk all tagged monitors
				bw=fr:GetBarWidth(); if (bw>width) then					-- find widest monitor
					width=bw; master=fr;								-- set alignment master
				end
				fr:SetMode(fr.storage.mode);							-- re-initialise to get fresh layout
			end end
			for _,fr in ipairs(mx.gui) do if (fr.group) then				-- walk all tagged monitors
				fr.FullWidth=width;
			end end
			master.alignColumn=true;									-- flag it for realignment
--_debug(floor(width));
			mon:UntagFrameGroup("group");								-- remove tag
		end
	end
end

function mx.api.Remove(frame)
	frame:Hide();
	for i,reg in ipairs(dccs.Monitor) do if (reg==frame.storage) then table.remove(dccs.Monitor,i); return; end end
end

function mx.api._PreClick(frame,LR,down)
end

function mx.api._OnClick(frame,LR,down)
	local setmaxpct=function(b,c,e)
		local e=floor(tonumber(e) or 100);
		if (e<5) then e=5; end
		frame.storage.max=e;
		frame.Ticker=0;
		frame:Regroup();
	end
	local addmonitor=function(ofr,count,ox,oy)
		repeat
			local fr=mx.api:CreateGUI();							-- create a new monitor
			fr.storage.freemove=ofr.storage.freemove;					-- copy lock
			fr:PositionMonitor(ofr.storage.left+ox,ofr.storage.bottom+oy);	-- place it as requested
			fr:Snap();													-- snap in place
			ofr=fr;														-- set new monitor as original source for sequential creation
			count=count-1;												-- one down, maybe more to go
		until(count<1);
	end
	if (LR=="RightButton") then
		local menu=DMMenuCreate(frame);
		if (frame.storage.locked) then
			menu:Button("Unlock",function() for _,fr in ipairs(mx.gui) do fr.storage.locked=nil; end end);	-- also do invisible ones
		else
			local subm;
			subm=menu:Submenu("Bar scale");
				subm:Radiobutton("Max percent:",setmaxpct,frame.storage.max~=0,nil,nil,nil,nil,1):AddEdit(((frame.storage.max==0 and 100) or frame.storage.max),40,"%",5,999);
				subm:Radiobutton("Automatic",function() frame.storage.max=0; frame.Ticker=0; frame:Regroup(); end,frame.storage.max==0,nil,nil,nil,nil,1);
			subm=menu:Submenu("Layout");
				subm:Button("Default",function() frame:SetMode("default"); frame:Regroup(); end);
				subm:Button("Compact",function() frame:SetMode("compact"); frame:Regroup(); end);
			subm=menu:Submenu("Add monitor");
				subm:Button("Below",function(b,c,e) addmonitor(frame,e,0,0-frame:GetHeight()); end):AddEdit(1,25,"",1,10,true);
				subm:Button("Above",function(b,c,e) addmonitor(frame,e,0,frame:GetHeight()); end):AddEdit(1,25,"",1,10,true);
				subm:Button("Right",function(b,c,e) addmonitor(frame,e,frame:GetWidth(),0); end):AddEdit(1,25,"",1,10,true);
				subm:Button("Left",function(b,c,e) addmonitor(frame,e,0-frame:GetWidth(),0); end):AddEdit(1,25,"",1,10,true);
			menu:Button("Lock",function() for _,fr in ipairs(mx.gui) do if (fr:IsVisible()) then fr.storage.locked=true; end end end);
			menu:Button("Lock this only",function() frame.storage.locked=true; end);
			menu:Checkbox("Glue sides",function(b,c) frame.storage.freemove=c; end,not frame.storage.freemove);
			menu:Checkbox("Auto-assign item",function(b,c) frame.storage.autoitem=true; end,frame.storage.autoitem);
			menu:Checkbox("Show items per minute",function(b,c) frame.storage.dropspeed=true; end,frame.storage.dropspeed);
			menu:Checkbox("Set timed goal",function(b,c,e) if (c) then frame.storage.goal=e; else frame.storage.goal=nil; end end,frame.storage.goal):AddEdit(frame.storage.goal or 20,45,"",1,1000);
			menu:Separator();
			menu:Button("Remove",function() frame:Remove(); end);
		end
		menu:Show();
	elseif (LR=="LeftButton") then
		frame.startL=frame:GetLeft();
		frame.startB=frame:GetBottom();
		for _,fr in ipairs(mx.gui) do
			fr.osL=frame.startL-fr:GetLeft();
			fr.osB=frame.startB-fr:GetBottom();
		end
	end
end

function mx.api.GetBarWidth(frame)
	return frame._UI.Header:GetWidth();
--	return frame._UI.Header:GetTextWidth();
end

function mx.api.SetPercent(frame,value)
	value=value or 0;
	if (value<0) then frame._UI.Counter:SetText("Quest"); frame._UI.Bar:SetWidth(0); return; end
	local fmt=(value<10 and "%.1f") or "%.0f";
	if (frame.storage.mode~="compact") then
		fmt=fmt.."%%";
		if (frame.storage.dropspeed) then
		end
	end
	frame._UI.Counter:SetText(fmt:format(value));
	frame.Ticker=CONST.GUIREFRESH;
	local thismax=1;
	if (frame.FollowItem) then
		if (value>(DropCount.Tracker.MaxSessionValue[frame.FollowItem] or -10)) then DropCount.Tracker.MaxSessionValue[frame.FollowItem]=value; end
		thismax=DropCount.Tracker.MaxSessionValue[frame.FollowItem];	-- assume auto-max
	end
	if (frame.storage.max>0) then thismax=frame.storage.max; end		-- set fixed
	if (thismax<1) then thismax=1; end									-- low clip
	if (value>thismax) then value=thismax; end							-- high clip
	if (value<0.01) then value=0.01; end

	frame._UI.Bar:SetWidth((frame.FullWidth/thismax)*value);

--	local part=value/frame.MaxSessionValue;
--	frame._UI.Bar:SetTexture("Interface\\AddOns\\LootCount_DropCount\\white.tga");
--	frame._UI.Bar:SetVertexColor(1-part,part,.4,.25);		-- r,g,b,a
--_debug(1-part,part,.4);
end

function mx.api._OnUpdate(frame,elapsed)
	if (frame.startL) then	-- moving
		mx.api.offsetL=frame:GetLeft()-frame.startL;
		mx.api.offsetB=frame:GetBottom()-frame.startB;
	end
	if (frame.glued) then	-- following one that moves
		frame:ClearAllPoints();
		frame:SetPoint("BOTTOMLEFT",UIParent,"BOTTOMLEFT",frame.storage.left+mx.api.offsetL or (UIParent:GetWidth()/2),frame.storage.bottom+mx.api.offsetB or (UIParent:GetHeight()*0.75));
	end
	if (frame.alignColumn) then											-- frame is master for column realignment
		if (frame:TagFrameGroup("group",true)) then						-- locate this monitor's vertical group members
			frame:UntagFrameGroup("group");
		end
		frame.alignColumn=nil;
	end
	if (frame.ItemNotInCache) then frame:Set[frame.storage.dropped[1](); return; end
	if (not frame.cbDelay) then return; end								-- no delay set - no update attempted
	frame.Ticker=frame.Ticker-elapsed; if (frame.Ticker>0) then return; end
	frame.Ticker=frame.Ticker+frame.cbDelay;
	local now=debugprofilestop();
--	frame:SetPercent((DropCount:TimedQueueRatio(getid(frame.FollowItem),true) or 0)*100);		-- item,noyield
	if (frame.cbUpdateSelf) then frame.cbUpdate(frame.cbUpdateSelf); elseif (frame.cbUpdate) then frame.cbUpdate(); end
	now=debugprofilestop()-now;
	if (now>10) then _debug("SetPercent:",string.format("%.2fms",now)); end
end
